import copy
import time

from pyomo.core import *
from pyomo.dae import *
from pyomo.environ import *

from kipet.core_methods.Optimizer import *


class VarianceEstimator(Optimizer):
    """Optimizer for variance estimation.

    Attributes:

        model (model): Pyomo model.

    """
    def __init__(self, model):
        super(VarianceEstimator, self).__init__(model)
        self.add_warm_start_suffixes(self.model)

        if not self._spectra_given:
            raise NotImplementedError("Variance estimator requires spectral data in model as model.D[ti,lj]")
        self._is_D_deriv = False

    def run_sim(self, solver, **kwds):
        raise NotImplementedError("VarianceEstimator object does not have run_sim method. Call run_opt")

    def run_opt(self, solver, **kwds):

        """Solves estimation following either the original Chen etal (2016) procedure or via the 
        maximum likelihood estimation with unknown covariance matrix. Chen's method solves a sequence 
        of optimization problems to determine variances and initial guesses for parameter estimation.
        The maximum likelihood estimation with unknown covariance matrix also solves a sequence of optimization
        problems is a more robust and reliable method, albeit somputationally costly.

        Args:

            solver_opts (dict, optional): options passed to the nonlinear solver
            
            method (str, optional): default is "Chen", other options are "max_likelihood" and "iterative_method"
            
            initial_sigmas (dict, optional): Required for "iterative_method"

            tee (bool,optional): flag to tell the optimizer whether to stream output
            to the terminal or not

            norm (optional): norm for checking convergence. The default value is the infinity norm,
            it uses same options as scipy.linalg.norm

            max_iter (int,optional): maximum number of iterations for Weifengs procedure. Default 400.

            tolerance (float,optional): Tolerance for termination by the change Z. Default 5.0e-5

            subset_lambdas (array_like,optional): Set of wavelengths to used in initialization problem 
            (Weifeng paper). Default all wavelengths.

            lsq_ipopt (bool,optional): Determines whether to use ipopt for solving the least squares 
            problems in Weifengs procedure. Default False. The default used scipy.least_squares.

            init_C (DataFrame,optional): Dataframe with concentration data used to start Weifengs procedure.

            report_time (bool, optional): True if variance estimation is timed. Default False
            
            fixed_device_variance (float, optional): If the device variance is known in advanced we can fix it here.
                                                Only to be used in conjunction with lsq_ipopt = True.
                                                
            secant_point (float, optional): Provides the second point (in addition to the init_sigma) for the secant method
                                            if not provided it will use 10 times the initial point
            
            device_range (tuple, optional): when using direct sigmas approach this provides the search region for delta
            
            num_points (int, optional): This provides the number of evaluations when using direct sigmas approach
            
            init_sigmas (float, optional): this will provide an initial value for the sigmas when using the alternative variance method
            
            individual_species (bool, optional): bool to see whether the overall model variance is returned or if the individual
                                                species' variances are returned based on the obtainde value for delta
            
            with_plots (bool, optional): if with_plots is provided, it will plot the log objective versus the
                                            iterations for the direct_sigmas method.

        Returns:

            Results from the optimization (pyomo model)

        """
        run_opt_kwargs = copy.copy(kwds)
        
        solver_opts = kwds.pop('solver_opts', dict())
        sigma_sq = kwds.pop('variances', dict())
        init_sigmas = kwds.pop('initial_sigmas', float())
        tee = kwds.pop('tee', False)
        method = kwds.pop('method', str())
        norm_order = kwds.pop('norm',np.inf)
        max_iter = kwds.pop('max_iter', 400)
        tol = kwds.pop('tolerance', 5.0e-5)
        A = kwds.pop('subset_lambdas', None)
        lsq_ipopt = kwds.pop('lsq_ipopt', False)
        init_C = kwds.pop('init_C', None)
        report_time = kwds.pop('report_time', False)
        individual_species = kwds.pop('individual_species', False)

        # additional arguments for inputs CS
        inputs = kwds.pop("inputs", None)
        inputs_sub = kwds.pop("inputs_sub", None)
        trajectories = kwds.pop("trajectories", None)
        fixedtraj = kwds.pop('fixedtraj', False)
        fixedy = kwds.pop('fixedy', False)
        yfix = kwds.pop("yfix", None)
        yfixtraj = kwds.pop("yfixtraj", None)

        jump = kwds.pop("jump", False)
        var_dic = kwds.pop("jump_states", None)
        jump_times = kwds.pop("jump_times", None)
        feed_times = kwds.pop("feed_times", None)

        species_list = kwds.pop('subset_components', None)
        
        # Modified variance estimation procedures arguments
        fixed_device_var = kwds.pop('fixed_device_variance', None)
        device_range = kwds.pop('device_range', None)
        secant_point2 = kwds.pop('secant_point', None)
        num_points = kwds.pop('num_points', None)
        with_plots = kwds.pop('with_plots', False)
        
        methods_available = ['originalchenetal', "direct_sigmas", "alternate"]
        
        if method not in methods_available:
            method = 'originalchenetal'
            print(f'Default method of {method} selected')
            print(f'The available methods are {", ".join(methods_available)}')
            
        if not self.model.alltime.get_discretization_info():
            raise RuntimeError('apply discretization first before initializing')

        self._create_tmp_outputs()
        
        objectives_map = self.model.component_map(ctype=Objective, active=True)
        active_objectives_names = []
        for obj in objectives_map.values():
            active_objectives_names.append(obj.cname())
            obj.deactivate()
            
        list_components = []
        if species_list is None:
            list_components = [k for k in self._mixture_components]
        else:
            for k in species_list:
                if k in self._mixture_components:
                    list_components.append(k)
                else:
                    warnings.warn("Ignored {} since is not a mixture component of the model".format(k))

        # Set up the correct components based on absorption
        if hasattr(self.model, 'non_absorbing'):
            warnings.warn("Overriden by non_absorbing")
            list_components = [k for k in self._mixture_components if k not in self._non_absorbing]
        
        if hasattr(self.model, 'known_absorbance'):
            warnings.warn("Overriden by species with known absorbance")
            list_components = [k for k in self._mixture_components if k not in self._known_absorbance]
        
        self._sublist_components = list_components
        
        if hasattr(self, '_abs_components'):
            self.component_set = self._abs_components
            self.component_var = 'Cs'
        else:
            self.component_set = self._sublist_components
            self.component_var = 'C'
        
        # Fixed imputs and trajectory section
        if inputs_sub is not None:
            from kipet.common.additional_inputs import add_inputs
            
            add_kwargs = dict(
                fixedtraj = fixedtraj,
                fixedy = fixedy, 
                inputs_sub = inputs_sub,
                yfix = yfix,
                yfixtraj = yfixtraj,
                trajectories = trajectories,
            )
            
            add_inputs(self, add_kwargs)

        # Dosing
        if jump:
            self.set_up_jumps(run_opt_kwargs)
        
        if report_time:
            start = time.time()
             
        # Call the chosen method
        if method == 'originalchenetal':  
            from kipet.variance_methods.chen_method import run_method
            results = run_method(self, solver, run_opt_kwargs)
            
        elif method == 'alternate':
            from kipet.variance_methods.alternate_method import run_alternate_method
            solver = 'ipopt'
            results = run_alternate_method(self, solver, run_opt_kwargs)
        
        elif method == 'direct_sigmas':
            from kipet.variance_methods.alternate_method import run_direct_sigmas_method
            results = run_direct_sigmas_method(self, solver, run_opt_kwargs)
            
        # Report time
        if report_time:
            end = time.time()
            print("Total execution time in seconds for variance estimation:", end - start)
        
        return results

    def _warn_if_D_negative(self):
    
        for t in self._meas_times:
            for l in self._meas_lambdas:
                if self.model.D[t, l] >= 0:
                    pass
                else:
                    self._is_D_deriv = True
        if self._is_D_deriv == True:
            print("Warning! Since D-matrix contains negative values Kipet is relaxing non-negativity on S")

        return None

    def _create_tmp_outputs(self):
        """Creates temporary files for loging solutions of each optimization problem

           This method is not intended to be used by users directly

        Args:

        Returns:
            None

        """
        self._tmp2 = "tmp_Z"
        self._tmp3 = "tmp_S"
        self._tmp4 = "tmp_C"
        
        with open(self._tmp2,'w') as f:
            f.write("temporary file for ipopt output")
        
        with open(self._tmp3,'w') as f:
            f.write("temporary file for ipopt output")
        
        with open(self._tmp4,'w') as f:
            f.write("temporary file for ipopt output")
            
    def solve_max_device_variance(self, solver, **kwds):
        """Solves the maximum likelihood formulation with (C-Z) = 0. solves ntp/2*log(eTe/ntp)
        and then calculates delta from the solution. See documentation for more details.
    
        Args:
            solver_opts (dict, optional): options passed to the nonlinear solver
        
            tee (bool,optional): flag to tell the optimizer whether to stream output
            to the terminal or not
        
            subset_lambdas (array_like,optional): Set of wavelengths to used in initialization problem 
            (Weifeng paper). Default all wavelengths.
    
        Returns:
    
            delta_sq (float): value of the max device variance
    
        """   
        solver_opts = kwds.pop('solver_opts', dict())
        tee = kwds.pop('tee', True)
        set_A = kwds.pop('subset_lambdas', list())
    
        if not set_A:
            set_A = self._meas_lambdas
    
        list_components = [k for k in self._mixture_components]
                    
        if hasattr(self.model, 'non_absorbing'):
            warnings.warn("Overriden by non_absorbing")
            list_components = [k for k in self._mixture_components if k not in self._non_absorbing]
        
        if hasattr(self.model, 'known_absorbance'):
            warnings.warn("Overriden by species with known absorbance")
            list_components = [k for k in self._mixture_components if k not in self._known_absorbance]
        
        self._sublist_components = list_components
            
        print("Solving For the worst possible device variance\n")
        
        self._warn_if_D_negative()
        
        obj = 0.0
        ntp = len(self._meas_times)
        nwp = len(self._meas_lambdas)
        
        if hasattr(self, '_abs_components'):
            self.component_set = self._abs_components
        else:
            self.component_set = self._sublist_components
        
        for t in self._meas_times:
            for l in set_A:
                D_bar = sum(self.model.Z[t, k] * self.model.S[l, k] for k in self.component_set)
                obj += (self.model.D[t, l] - D_bar)**2
       
        self.model.init_objective = Objective(expr=obj)
        
        opt = SolverFactory(solver)
    
        for key, val in solver_opts.items():
            opt.options[key]=val
        solver_results = opt.solve(self.model,
                                   tee=tee)
        
        print("values for the parameters in case with no model variance")
        for k, v in self.model.P.items():
            print(k, v.value)
        
        etaTeta = 0
        for t in self._meas_times:
            for l in set_A:
                D_bar = sum(value(self.model.Z[t, k]) * value(self.model.S[l, k]) for k in self.component_set)
                etaTeta += (value(self.model.D[t, l])- D_bar)**2
    
        deltasq = etaTeta/(ntp*nwp)
        
        self.model.del_component('init_objective')
        print("worst case delta squared: ", deltasq)
    
        return deltasq
    
    def solve_sigma_given_delta(self, solver, **kwds):
        """Function that solves for model variances based on a given device variance. Solves
        The log likelihood function and returns a dictionary containg all sigmas, including
        the device/delta in order to easily apply to the parameter estimation problem.
           This method is intended to be used by users directly
        Args:
            solver (str): solver to use to solve the problems (recommended "ipopt")
            
            delta (float): the device variance squared 
        
            tee (bool,optional): flag to tell the optimizer whether to stream output
            to the terminal or not
        
            solver_opts (dict,optional): Dictionary containing solver options
        
            subset_lambdas (array_like,optional): Set of wavelengths to used in problem 
            (Weifeng paper). Default all wavelengths.
        Returns:
            all_variances (dict): dictionary containg all sigmas, including the device/delta
        """   
        from kipet.variance_methods.alternate_method import _solve_sigma_given_delta
        
        solver_opts = kwds.pop('solver_opts', dict())
        tee = kwds.pop('tee', True)
        set_A = kwds.pop('subset_lambdas', list())
        delta = kwds.pop('delta', dict())
        
        residuals, sigma_vals, stop_it, results = _solve_sigma_given_delta(self, 
                                                                           solver, 
                                                                           subset_lambdas= set_A, 
                                                                           solver_opts = solver_opts, 
                                                                           tee=tee, 
                                                                           delta = delta)
        all_variances = sigma_vals
        all_variances['device'] = delta
        return all_variances


"""
Data Handling for Kipet

Created on Mon Aug 24 04:01:57 2020

@author: kevin
"""
import inspect
import os

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from plotly import __version__
from plotly.offline import download_plotlyjs, init_notebook_mode, plot, iplot

from kipet.core_methods.data_tools import *
from kipet.visuals.plots import colors

data_categories = ['concentration', 'spectral', 'state', 'trajectory', 'custom']

class DataBlock():
    
    def __init__(self):
        
        self.datasets = {}
        
    def __getitem__(self, value):
        
        return self.datasets[value]
         
    def __str__(self):
        
        format_string = "{:<20}{:<15}{:<30}{:<15}{:<30}\n"
        data_str = 'DataBlock:\n'
        data_str += format_string.format(*['Name', 'Category', 'Components', 'Size', 'File'])
        
        for dataset in self.datasets.values():
            
            species_str = ', '.join([s for s in dataset.species[:3]])
            if len(dataset.species) > 3:
                species_str += f', ...'
            
            if dataset.file is not None:
                data_str += format_string.format(f'{dataset.name}', f'{dataset.category}', f'{species_str}', f'{dataset.data.shape}', f'{dataset.file.name}')
            else:
                data_str += format_string.format(f'{dataset.name}', f'{dataset.category}', f'{species_str}', f'{dataset.data.shape}', 'None')
                
        
        return data_str

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        for param, data in self.datasets.items():
            yield data
            
    def __len__(self):
        return len(self.datasets)
    
            
    def add_dataset(self, *args, **kwargs):
        
        """Should handle a series of different input methods:
          
        KP = KineticParameter('k1', init=1.0, bounds=(0.01, 10))
        builder.add_parameter_temp(KP)
        
        - or -
        
        builder.add_parameter('k1', init=1.0, bounds=(0.01, 10))
        
        - or -
        
        builder.add_parameter('k1', 1.0, (0.01, 10))
            
        """        
        category = kwargs.pop('category', None)
        data = kwargs.pop('data', None)
        file = kwargs.pop('file', None)
        
        units = None
        notes = None
        description = None
        
        if isinstance(args[0], DataSet):
            self.datasets[args[0].name] = args[0]
        
        else:
            if isinstance(args[0], str):
                
                dataset = DataSet(
                    args[0], 
                    category=category, 
                    data=data, 
                    units=units, 
                    notes=notes, 
                    file=file,
                    description=description,
                 )
                self.datasets[args[0]] = dataset
    
        return None
    
    def _check_duplicates(self):
        """Easy data mapping to ElementBlocks"""

        blocks = ['components', 'states', 'algebraics']
        
        all_data_cols = []
        for dataset in self.datasets.values():
            if dataset.category in ['spectral', 'trajectory']:
                continue
            all_data_cols.extend(dataset.data.columns)
        
        num_duplicates = len(all_data_cols) - len(set(all_data_cols))
        if num_duplicates != 0:
            raise ValueError('Duplicates ({num_duplicates}) in component, state, and algebraic data detected!')

    @property
    def names(self):
        return [dataset.name for dataset in self.datasets.values()]
        

    @property
    def time_span(self):
        
        start_time = min([dataset.time_span[0] for dataset in self.datasets.values()])
        stop_time = max([dataset.time_span[1] for dataset in self.datasets.values()])

        return (start_time, stop_time)

class DataSet(object):
    
    """The specific data object"""
    
    def __init__(self, 
                 name, 
                 category=None, 
                 data=None, 
                 units=None, 
                 notes=None, 
                 file=None,
                 description=None,
                 ):
        
        self.name = name
        self.category = category
        # if category not in data_categories:
        #     raise ValueError(f'The data type must be in one of the following: {", ".join(data_categories)}')

        self.data = data
        self.file = file
        self.units = units
        self.notes = notes
        self.description = description

        self._check_input()

    def __repr__(self):
        
        if self.data is not None:
            return f'DataSet(name={self.name}, category={self.category}, data={self.data.shape})'
        else:
            return f'DataSet(name={self.name}, category={self.category}, data=None)'
    def __str__(self):
        
        if self.data is not None:
            return f'DataSet(name={self.name}, category={self.category}, data={self.data.shape})'
        else:
            return f'DataSet(name={self.name}, category={self.category}, data=None)'

    def _check_input(self):
        """Checks whether the data has been provided as a pandas dataframe or
        if a filename has been provided. If not, the DataBlock is initialized
        without any data
        """
        if self.data is not None:
            if not isinstance(self.data, pd.DataFrame):
                raise ValueError('Data must be a pandas DataFrame instance')
            
        elif self.data is None and self.file is not None:
            # if self.category == 'spectral':
            #     self.data = read_file
            print(self.file)
            self.data = read_file(self.file)
            
        else:
            print('DataBlock object initialized without data')
    
        return None
    
    def remove_negatives(self):
        """Replaces the negative values with zero"""
        self.data[self.data < 0] = 0
        return None
    
    def show_data(self):
        """Method to show the data using Plotly"""
        
        if self.data is None:
            print('No data to plot')
            return None
        
        if self.category in ['concentration', 'state', 'trajectory', 'custom']:
            self._plot_2D_data()
            
        if self.category == 'spectral':
            self._plot_spectral_data()
    
        return None
    
    def _plot_2D_data(self):
        """Simple plots for showing concentration or complementary state data
        
        """
        y_axis_text = 'Measured Value'
        if self.description is not None:
            y_axis_text = self.description
            
        if self.units is not None:
            y_axis_text = ' '.join([y_axis_text, '[' + self.units[0] +']'])
            
        x_axis_text = 'Time'
        
        if self.units is not None and len(self.units) == 2:
            x_axis_text = ' '.join([x_axis_text, '[' + self.units[1] +']'])
        
        fig = go.Figure()
        for i, cols in enumerate(self.data.columns):
            fig.add_trace(go.Scatter(x=self.data.index,
                                     y=self.data[cols],
                                     mode='markers',
                                     name=cols,
                                     marker=dict(size=10, opacity=0.5, color=colors[i])),
                          )
           
        fig.update_layout(title_text=f'{self.category.capitalize()} data: {self.name}',
                          xaxis_title=f'{x_axis_text}',
                          yaxis_title=f'{y_axis_text}',
                          title_font_size=30)

        plot(fig)

        return None

    def _plot_spectral_data(self, dimension='3D'):
        """ Plots spectral data
        
            Args:
                dataFrame (DataFrame): spectral data
              
            Returns:
                None
    
        """
        if dimension=='3D':
            
            fig = go.Figure()
            fig.add_trace(go.Surface(x=self.data.columns,
                               y=self.data.index,
                               z=self.data.values,
                              ))
            
            fig.update_layout(scene = dict(
                                xaxis_title='Wavelength',
                                yaxis_title='Time',
                                zaxis_title='Spectra'),
                                margin=dict(r=100, b=50, l=100, t=50),
                                title_text=f'{self.name}: {self.category.capitalize()} data',
                                title_font_size=30)
            
            plot(fig)
            
            return None

    # Properties
    @property
    def shape(self):
        if self.data is not None:
            return self.data.shape
        else:
            return None

    @property
    def size(self):
        if self.data is not None:
            return len(self.data.columns)
        else:
            return None
    
    @property
    def species(self):
        if self.data is not None and self.category in ['concentration', 'state', 'trajectory', 'custom']:
            return list(self.data.columns)
        elif self.data is None:
            return []
        else:
            return []
        
    @property
    def time_span(self):
        return self.data.index.min(), self.data.index.max()


if __name__ == '__main__':

    pass


"""Multiple Experiment Estimator"""

# Standard library imports
import copy
import math
import os
import scipy.stats as st

# Third party imports
import numpy as np
import pandas as pd
from pyomo import *
from pyomo.dae import *
from pyomo.environ import *
from scipy.sparse import coo_matrix

# KIPET library imports
from kipet.common.parameter_handling import initialize_parameters
from kipet.common.objectives import conc_objective, comp_objective, absorption_objective
from kipet.common.read_hessian import split_sipopt_string
from kipet.mixins.PEMixins import PEMixins
from kipet.core_methods.fe_factory import *
from kipet.core_methods.FESimulator import *
from kipet.core_methods.Optimizer import *
from kipet.core_methods.ParameterEstimator import *
from kipet.core_methods.PyomoSimulator import *
from kipet.core_methods.VarianceEstimator import *

from kipet.top_level.variable_names import VariableNames

__author__ = 'Michael Short, Kevin McBride'  #: February 2019 - October 2020

class MultipleExperimentsEstimator(PEMixins, object):
    """This class is for Estimation of Variances and parameters when we have multiple experimental datasets.
    This class relies heavily on the Pyomo block class as we put each experimental class into its own block.
    This blocks are first run individually in order to find good initializations and then they are linked and
    run together in a large optimization problem.

    Parameters
    ----------
    model : TemplateBuilder
        The full model TemplateBuilder problem needs to be fed into the MultipleExperimentsEstimator. This
        pyomo model will form the basis for all the optimization tasks
    """

    def __init__(self, reaction_models):
        
        #super(MultipleExperimentsEstimator, self).__init__()
        self.reaction_models = reaction_models
        self.experiments = list(self.reaction_models.keys())
        self._idx_to_variable = dict()
        
        self.variances = {name: model.variances for name, model in self.reaction_models.items()}
        self.make_sublist()
        
        self._n_meas_times = 0
        self._n_meas_lambdas = 0
        self._n_actual = 0
        self._n_params = 0
        
        self._spectra_given = True
        self._concentration_given = False
        
        self.global_params = None
        self.parameter_means = False
        
        self.__var = VariableNames()
  
    
    def make_sublist(self):

        self._sublist_components = {}        
        for name, model in self.reaction_models.items():
            self._sublist_components[name] = [comp.name for comp in model.components if comp.state == 'concentration']
        return None
        
    def _define_reduce_hess_order_mult(self):
        """This function is used to link the variables to the columns in the reduced
            hessian for multiple experiments.   
           
            Not meant to be used directly by users
        """
        self.model.red_hessian = Suffix(direction=Suffix.IMPORT_EXPORT)
        count_vars = 1
        model_obj = self.model.experiment

        for i in self.experiments:
            if hasattr(self.reaction_models[i].model, 'C'):
            #if self._spectra_given:
                if hasattr(self, 'model_variance') and self.model_variance or not hasattr(self, 'model_variance'):
                    count_vars = self._set_up_reduced_hessian(model_obj[i], self.model.experiment[i].meas_times, self._sublist_components[i], 'C', count_vars)
        
        for i in self.experiments:
            if hasattr(self.reaction_models[i].model, 'S'):
            #if self._spectra_given:
                if hasattr(self, 'model_variance') and self.model_variance or not hasattr(self, 'model_variance'):       
                    count_vars = self._set_up_reduced_hessian(model_obj[i], self.model.experiment[i].meas_lambdas, self._sublist_components[i], 'S', count_vars)
                
        for i in self.experiments:
            for v in model_obj[i].P.values():
                if v.is_fixed():
                    print(v, end='\t')
                    print("is fixed")
                    continue
                self._idx_to_variable[count_vars] = v
                self.model.red_hessian[v] = count_vars
                count_vars += 1
        return None
           
    def _set_up_marks(self, conc_only=True):
        """Set up the data based on the number of species and measurements
        
        """    
        nt = np.cumsum(np.array([len(self.model.experiment[exp].meas_times) for i, exp in enumerate(self.experiments)]))
        self.t_mark = {i: n for i, n in enumerate(nt)}
        nt = nt[-1]
        self._n_meas_times = nt
        
        #nc = np.cumsum(np.array([len(self._sublist_components[exp]) for i, exp in enumerate(self.experiments)]))
        nc = np.cumsum([len(self.reaction_models[k].components) for k in self.reaction_models.keys()])
        
        self.n_mark = {i: n for i, n in enumerate(nc)}
        nc = nc[-1]
        self._n_actual = nc
        
        #nparams = np.cumsum(np.array([self._get_nparams(self.model.experiment[exp]) for i, exp in enumerate(self.experiments)]))
        nparams = np.cumsum([len(self.reaction_models[k].parameters) for k in self.reaction_models.keys()])
         
        self.p_mark = {i: n for i, n in enumerate(nparams)}
        nparams = nparams[-1]
        self._n_params = nparams
        
        if not conc_only:
            
            nw = np.cumsum(np.array([len(self.model.experiment[exp].meas_lambdas) for i, exp in enumerate(self.experiments)]))
            self.l_mark = {i: n for i, n in enumerate(nw)}
            nw = nw[-1]
            self._n_meas_lambdas = nw
            
        else:
            self.l_mark = {}
            self._n_meas_lambdas = 0
        
        print(self.t_mark, self.n_mark, self.p_mark, self.l_mark)
        print(self._n_meas_times, self._n_actual, self._n_params, self._n_meas_lambdas)
        
        return None
    
    def _display_covariance(self, variances_p):
        """Displays the covariance results to the console
        """
        #print(self.confidence_interval)
        number_of_stds = 1#st.norm.ppf(1-(1-self.confidence_interval)/2)
        
        print(number_of_stds)
        
        print('\nParameters:')
        for exp in self.experiments:
            for k, p in self.model.experiment[exp].P.items():
                if p.is_fixed():
                    continue
                print('{}, {}'.format(k, p.value))
        print('\nConfidence intervals:')
        for exp in self.experiments:
            for i, (k, p) in enumerate(self.model.experiment[exp].P.items()):
                if p.is_fixed():
                    continue
                std = (self.variance_scale*variances_p[i])** 0.5
                print('{} ({},{})'.format(k, p.value - number_of_stds*std, p.value + number_of_stds*std))
                
        if hasattr(self.model.experiment[exp], 'Pinit'):
            print('\nLocal Parameters:')
            for exp in self.experiments:
                for k in self.model.experiment[exp].Pinit.keys():
                    self.model.experiment[exp].Pinit[k] = self.model.experiment[exp].init_conditions[k].value
                    print('{}, {}'.format(k, self.model.experiment[exp].Pinit[k].value))
            print('\nConfidence intervals:')
            for exp in self.experiments:
                for i, k in enumerate(self.model.experiment[exp].Pinit.keys()):
                    self.model.experiment[exp].Pinit[k] = self.model.experiment[exp].init_conditions[k].value
                    #std = (self.variance_scale*variances_p[i])** 0.5
                    print('{} ({},{})'.format(k, 
                                              self.model.experiment[exp].Pinit[k].value - (number_of_stds*variances_p[i]**0.5), 
                                              self.model.experiment[exp].Pinit[k].value + (number_of_stds*variances_p[i]**0.5)))
                            
        return None
        
    def _compute_covariance(self, hessian, variances):
        """
        Computes the covariance matrix for the paramaters taking in the Hessian
        matrix and the variances.
        Outputs the parameter confidence intervals.

        This function is not intended to be used by the users directly

        """
        if not self.spectra_problem:
            self._set_up_marks()
            res = {}
            for exp in self.experiments:
                res.update(self._compute_residuals(self.model.experiment[exp], exp_index=exp))

            H = hessian[-self._n_params:, :]
            variances_p = np.diag(H)

        else:         
            self._set_up_marks(conc_only=False) 
            variances_p, covariances_p = self._variances_p_calc(hessian, variances)   
        
        print(f'VP: {variances_p}')
        self._display_covariance(variances_p)
        
        return None
        
    def _compute_B_matrix(self, variances, **kwds):
        """Builds B matrix for calculation of covariances

            This method is not intended to be used by users directly

        Args:
            variances (dict): variances

        Returns:
            None
        """
        nt = self._n_meas_times
        nw = self._n_meas_lambdas
        nc = self._n_actual
        nparams = self._n_params

        npn = np.r_[self.n_mark[0], np.diff(list(self.n_mark.values()))]
        npt = np.r_[self.t_mark[0], np.diff(list(self.t_mark.values()))]
        npl = np.r_[self.l_mark[0], np.diff(list(self.l_mark.values()))]
        npp = np.r_[self.p_mark[0], np.diff(list(self.p_mark.values()))]
        exp_lookup = {i: exp for i, exp in enumerate(self.experiments)}
        ntheta = sum(npn*(npt + npl) + npp)
        
        exp_count = 0
        timeshift = 0 
        waveshift = 0
        
        rows = []
        cols = []
        data = []
        
        meas_times = {exp : {indx: time for indx, time in enumerate(self.model.experiment[exp].meas_times)} for exp in self.experiments}
        meas_lambdas = {exp : {indx: wave for indx, wave in enumerate(self.model.experiment[exp].meas_lambdas)} for exp in self.experiments}
        
        for i in range(nt):
            for j in range(nw):
            
                nc = npn[exp_count]
                if i == self.t_mark[exp_count] and j == self.l_mark[exp_count]:
                    exp_count += 1
                    timeshift = i
                    waveshift = j    
           
                exp = exp_lookup[exp_count]
                
                for comp_num, comp in enumerate(self._sublist_components[exp]):
                    
                    if i - timeshift in list(range(npt[exp_count])):
                        time = meas_times[exp][i - timeshift]
                    
                    if j - waveshift in list(range(npl[exp_count])):
                        wave = meas_lambdas[exp][j - waveshift]
   
                    r_idx1 = i*nc + comp_num
                    r_idx2 = j*nc + comp_num + nc*nt
                    c_idx =  i*nw + j
                    
                    rows.append(r_idx1)
                    cols.append(c_idx)
                    data.append(-2 * self.model.experiment[exp].S[wave, comp].value / (self.variances[exp]['device']))
  
                    rows.append(r_idx2)
                    cols.append(c_idx)
                    data.append(-2 * self.model.experiment[exp].C[time, comp].value / (self.variances[exp]['device']))
                         
        B_matrix = coo_matrix((data, (rows, cols)), shape=(ntheta, nw * nt)).tocsr()
        self.B_matrix = B_matrix
        
        return B_matrix
        
    def _compute_Vd_matrix(self, variances, **kwds):
        """Builds d covariance matrix

            This method is not intended to be used by users directly

        Args:
            variances (dict): variances

        Returns:
            None
        """
        row = []
        col = []
        data = []
        nt = self._n_meas_times
        nw = self._n_meas_lambdas
        nc = self._n_actual

        v_array = np.zeros(nc)
        s_array = np.zeros(nw * nc)
        
        count = 0
        for x in self.experiments:
            for k, c in enumerate(self._sublist_components[x]):
                v_array[count] = variances[x][c]
                count += 1
        
        kshift = 0
        jshift = 0
        knum = 0
        jnum=0
        count=0
        exp_count = 0
        
        for x in self.experiments:
            kshift += knum
            jshift += jnum
            if exp_count != 0:
                kshift+=1

            for j, l in enumerate(self.model.experiment[x].meas_lambdas):
                for k, c in enumerate(self._sublist_components[x]):
                    
                    if exp_count == 0:
                        nc = self.n_mark[exp_count]
                        nt = self.t_mark[exp_count]
                        nw = self.l_mark[exp_count]
                    else: 
                        nt = (self.t_mark[exp_count] - self.t_mark[exp_count - 1] )
                        nw = (self.l_mark[exp_count] - self.l_mark[exp_count - 1])

                    s_array[(j+jshift) * nc + (k+kshift)] = self.model.experiment[x].S[l, c].value
                    idx = (j+jshift) * nc + (k+kshift)
                    knum = max(knum,k)
                    count += 1
                
                jnum = max(jnum,j)
            exp_count += 1

        row = []
        col = []
        data = []
        nt = self._n_meas_times
        nw = self._n_meas_lambdas
        nd = nt * nw
        v_device = list()
        
        for x in self.experiments:
            if 'device' in variances[x]:
                v_device.append(variances[x]['device']) 
        
        exp_count = 0
        v_device_exp = v_device[exp_count]
        for i in range(nt):
            for j in range(nw):
                if i == self.t_mark[exp_count] and j == self.l_mark[exp_count]:
                    exp_count += 1
                    v_device_exp = v_device[exp_count]
                    
                val = sum(v_array[k] * s_array[j * nc + k] ** 2 for k in range(nc)) + v_device_exp
                row.append(i * nw + j)
                col.append(i * nw + j)
                data.append(val)
                for p in range(nw):
                    if j != p:
                        val = sum(v_array[k] * s_array[j * nc + k] * s_array[p * nc + k] for k in range(nc))
                        row.append(i * nw + j)
                        col.append(i * nw + p)
                        data.append(val)

        Vd_matrix = coo_matrix((data, (row, col)), shape=(nd, nd)).tocsr()
        
        self.Vd_matrix = Vd_matrix
    
        return Vd_matrix
    
    """
    Save the sim results in self.sim_results as a dict: self._sim_solved set to True
    """
    
    def solve_hessian(self, solver, **kwargs):
        """Solves for the Hessian regardless of data source - only uses ipopt_sens
        
        Args:
            sigma_sq (dict): variances

            optimizer (SolverFactory): Pyomo Solver factory object

            tee (bool,optional): flag to tell the optimizer whether to stream output
            to the terminal or not

        Returns:
            hessian (np.ndarray): The hessian matrix for covariance calculations
            
        """
        tee = kwargs.pop('tee', True)
        solver_opts = kwargs.pop('solver_opts', dict())
        
        if solver == 'k_aug':
            solver = 'ipopt_sens'
        
        if solver == 'ipopt_sens':
            if not 'compute_red_hessian' in solver_opts.keys():
                solver_opts['compute_red_hessian'] = 'yes'
                
        # Create the optimizer
        optimizer = SolverFactory(solver)
        for key, val in solver_opts.items():
            optimizer.options[key] = val
            
        # Declare the model
        m = self.model
        self._define_reduce_hess_order_mult()
        
        # Solve
        self._tmpfile = "ipopt_hess"
        solver_results = optimizer.solve(m, 
                                          tee=tee,
                                          logfile=self._tmpfile,
                                          report_timing=True)
        
        with open(self._tmpfile, 'r') as f:
            output_string = f.read()
        if os.path.exists(self._tmpfile):
            os.remove(self._tmpfile)

        ipopt_output, hessian_output = split_sipopt_string(output_string)
        hessian = read_reduce_hessian(hessian_output, len(self._idx_to_variable))
        
        return hessian
        
    def _scale_variances(self,):
        """Option to scale the variances for MEE"""
        var_scaled = dict()
        for s,t in self.variances.items():
            maxx = max(list(t.values()))
            ind_var = dict()
            for i,j in t.items():
                ind_var[i] = j/maxx
            var_scaled[s] = ind_var
        self.variances = var_scaled
        self.variance_scale = maxx
        
        return None
    
    # def calculate_parameter_averages(self):
        
    #     p_dict = {}
    #     c_dict = {}
        
    #     for key, model in self.reaction_models.items():
    #         for param, obj in getattr(model, self.__var.model_parameter).items():
    #             if param not in p_dict:
    #                 p_dict[param] = obj.value
    #                 c_dict[param] = 1
    #             else:
    #                 p_dict[param] += obj.value
    #                 c_dict[param] += 1
                    
    #     self.avg_param = {param: p_dict[param]/c_dict[param] for param in p_dict.keys()}
        
    #     return None
        
    # def initialize_parameters(self):
        
    #     self.calculate_parameter_averages()
        
    #     for key, model in self.reaction_models.items():
    #         for param, obj in getattr(model, self.__var.model_parameter).items():
    #             obj.value = self.avg_param[param] 
                
    #     return None
            
    def solve_consolidated_model(self, 
                                 global_params=None,
                                 **kwargs):
        """This function consolidates the individual models into a single
        optimization problem that links the parameters and spectra (if able)
        from each experiment
        
        """
        solver_opts = kwargs.get('solver_opts', {'linear_solver': 'ma57'})
        tee = kwargs.get('tee', False)
        scaled_variance = kwargs.get('scaled_variance', False)
        shared_spectra = kwargs.get('shared_spectra', True)
        solver = kwargs.get('solver', 'ipopt')
        parameter_means = kwargs.get('mean_start', True)
        
        print("\nSOLVING PARAMETER ESTIMATION FOR MULTIPLE DATASETS\n")
       
        combined_model = ConcreteModel()
        
        self.variance_scale = 1
        if scaled_variance == True:
            self._scale_variances()
            
        # if parameter_means:
        #     initialize_parameters(self.reaction_models)
        
        if global_params is None:
            global_params = self.all_params
        
        def build_individual_blocks(m, exp):
            """This function forms the rule for the construction of the individual blocks 
            for multiple experiments, referenced in run_parameter_estimation. This function 
            is not meant to be used by users directly.
            
            Args:
                m (pyomo Concrete model): the concrete model that we are adding the block to
                exp (list): a list containing the experiments
                
            Returns:
                Pyomo model: Pyomo model inside the block
                
            """
            list_components = self.reaction_models[exp].components.names
            #WITH_D_VARS as far as I can tell should always be True, unless we ignore the device noise
            with_d_vars= True
            m = copy.copy(self.reaction_models[exp].p_model)
            
            # Quick fix - I don't know what is causing this
            if hasattr(m, 'alltime_domain'):
                m.del_component('alltime_domain')
            
            if with_d_vars and hasattr(m, 'D'): #self._spectra_given:
              
                m.D_bar = Var(m.meas_times,
                              m.meas_lambdas)
    
                def rule_D_bar(m, t, l):   
                    return m.D_bar[t, l] == sum(getattr(m, self.__var.concentration_spectra)[t, k] * getattr(m, self.__var.spectra_species)[l, k] for k in self.reaction_models[exp].p_estimator._sublist_components)
    
                m.D_bar_constraint = Constraint(m.meas_times,
                                                m.meas_lambdas,
                                                rule=rule_D_bar)
            
            m.error = Var(bounds = (0, None))
                
            def rule_objective(m):
                
                expr = 0
                spectral_term = 0
                concentration_term = 0
                measured_concentration_term = 0
                complementary_state_term = 0
                weights = [1, 1, 1, 1]
                obj_variances = self.variances
                
                if hasattr(m, self.__var.spectra_data):
                    spectral_term = absorption_objective(m, 
                                                 device_variance=obj_variances[exp]['device'],
                                                 g_option=self.reaction_models[exp]._G_data['G_contribution'],
                                                 with_d_vars=with_d_vars,
                                                 shared_spectra=shared_spectra,
                                                 species_list=list_components)
                    
                    concentration_term = conc_objective(m, variance=obj_variances[exp], source='spectra')
                
                if hasattr(m, self.__var.concentration_measured):
                    measured_concentration_term = conc_objective(m, variance=obj_variances[exp])
                
                if hasattr(m, self.__var.state):
                    complementary_state_term = comp_objective(m, variance=obj_variances[exp])
                    
                expr = weights[0]*spectral_term + \
                    weights[1]*concentration_term + \
                    weights[2]*measured_concentration_term + \
                    weights[3]*complementary_state_term
    
                return m.error == expr
    
            m.obj_const = Constraint(rule=rule_objective)
                
            return m  
        
        combined_model.experiment = Block(self.experiments, rule=build_individual_blocks)
        combined_model.map_exp_to_count = dict(enumerate(self.experiments))
        
        def param_linking_rule(m, exp, param):
            prev_exp = None
            key = next(key for key, value in combined_model.map_exp_to_count.items() if value == exp)
            if key == 0:
                return Constraint.Skip
            else:
                for key, val in combined_model.map_exp_to_count.items():
                    if val == exp:
                        prev_exp = combined_model.map_exp_to_count[key-1]
                if param in global_params and prev_exp != None:
                    return getattr(combined_model.experiment[exp], self.__var.model_parameter)[param] == getattr(combined_model.experiment[prev_exp], self.__var.model_parameter)[param]
                else:
                    return Constraint.Skip
        
        set_fixed_params=set()
        
        for exp in self.experiments:
            for param, param_obj in getattr(combined_model.experiment[exp], self.__var.model_parameter).items():
                if param_obj.is_fixed():
                    set_fixed_params.add(param)
                    
        print("Fixed parameters are: ", set_fixed_params)
        
        set_params_across_blocks = self.all_params.difference(set_fixed_params)
        combined_model.parameter_linking = Constraint(self.experiments, set_params_across_blocks, rule = param_linking_rule)
        
        def wavelength_linking_rule(m, exp, wave, comp):
            prev_exp = None
            key = next(key for key, value in combined_model.map_exp_to_count.items() if value == exp)
            if key == 0:
                return Constraint.Skip
            else:
                for key, val in combined_model.map_exp_to_count.items():
                    if val == exp:
                        prev_exp = combined_model.map_exp_to_count[key-1]
                if wave in self.all_wavelengths and prev_exp != None:
                    if comp in combined_model.experiment[prev_exp].mixture_components and comp in combined_model.experiment[exp].mixture_components:
                        return getattr(combined_model.experiment[exp], self.__var.spectra_species)[wave,comp] == getattr(combined_model.experiment[prev_exp], self.__var.spectra_species)[wave,comp]
                    else:
                        return Constraint.Skip
                else:
                    return Constraint.Skip

        if shared_spectra == True:
            combined_model.spectra_linking = Constraint(self.experiments, self.all_wavelengths, self.all_species, rule = wavelength_linking_rule)
        
        # Add in experimental weights
        combined_model.objective = Objective(sense=minimize, expr=sum(b.error for b in combined_model.experiment[:]))
        
        self.model = combined_model
        
        if solver in ['k_aug', 'ipopt_sens']:
            covariance = True
            hessian = self.solve_hessian(solver, tee=tee, solver_opts=solver_opts)
            self._compute_covariance(hessian, self.variances)
        else:
            optimizer = SolverFactory('ipopt')
            optimizer.solve(combined_model, options=solver_opts, tee=True)            
        
        solver_results = {}
        
        for i in combined_model.experiment:
            solver_results[i] = ResultsObject()
            solver_results[i].load_from_pyomo_model(combined_model.experiment[i])
                                            
        return solver_results
    
    @property
    def all_params(self):
        set_of_all_model_params = set()
        for name, model in self.reaction_models.items():
            set_of_all_model_params = set_of_all_model_params.union(model.parameters.names)
        return set_of_all_model_params
    
    
    @property
    def all_wavelengths(self):
        set_of_all_wavelengths = set()
        for name, model in self.reaction_models.items():
            set_of_all_wavelengths = set_of_all_wavelengths.union(list(model.model.meas_lambdas))
        return set_of_all_wavelengths
    
    @property
    def all_species(self):
        set_of_all_species = set()
        for name, model in self.reaction_models.items():
            set_of_all_species = set_of_all_species.union(model.components.names)
        return set_of_all_species
    
#%%

# combined_model = kipet_model.mee.model

# from kipet.core_methods.ResultsObject import ResultsObject

# for i in combined_model.experiment:
#       solver_results[i] = ResultsObject()
#       solver_results[i].load_from_pyomo_model(combined_model.experiment[i])
                                            
# print(solver_results)



"""
Module for reading and converting hessian data
"""
# Standard library imports
import contextlib
import os

# Thirs 
import numpy as np

def supress_stdout(func):
    def wrapper(*a, **ka):
        with open(os.devnull, 'w') as devnull:
            with contextlib.redirect_stdout(devnull):
                func(*a, **ka)
    return wrapper


def split_sipopt_string(output_string):
    start_hess = output_string.find('DenseSymMatrix')
    ipopt_string = output_string[:start_hess]
    hess_string = output_string[start_hess:]
    return (ipopt_string, hess_string)


def split_k_aug_string(output_string):
    start_hess = output_string.find('')
    ipopt_string = output_string[:start_hess]
    hess_string = output_string[start_hess:]
    return (ipopt_string, hess_string)


def read_reduce_hessian2(hessian_string, n_vars):
    hessian_string = re.sub('RedHessian unscaled\[', '', hessian_string)
    hessian_string = re.sub('\]=', ',', hessian_string)

    hessian = np.zeros((n_vars, n_vars))
    for i, line in enumerate(hessian_string.split('\n')):
        if i > 0:
            hess_line = line.split(',')
            if len(hess_line) == 3:
                row = int(hess_line[0])
                col = int(hess_line[1])
                hessian[row, col] = float(hess_line[2])
                hessian[col, row] = float(hess_line[2])
    return hessian


def read_reduce_hessian(hessian_string, n_vars):
    hessian = np.zeros((n_vars, n_vars))
    for i, line in enumerate(hessian_string.split('\n')):
        if i > 0:  # ignores header
            if line not in ['', ' ', '\t']:
                hess_line = line.split(']=')
                if len(hess_line) == 2:
                    value = float(hess_line[1])
                    column_line = hess_line[0].split(',')
                    col = int(column_line[1])
                    row_line = column_line[0].split('[')
                    row = int(row_line[1])
                    hessian[row, col] = float(value)
                    hessian[col, row] = float(value)
    return hessian


def read_reduce_hessian_k_aug(hessian_string, n_vars):
    hessian = np.zeros((n_vars, n_vars))
    for i, line in enumerate(hessian_string.split('\n')):
        if i > 0:  # ignores header
            if line not in ['', ' ', '\t']:
                hess_line = line.split(']=')
                if len(hess_line) == 2:
                    value = float(hess_line[1])
                    column_line = hess_line[0].split(',')
                    col = int(column_line[1])
                    row_line = column_line[0].split('[')
                    row = int(row_line[1])
                    hessian[row, col] = float(value)
                    hessian[col, row] = float(value)
    return hessian


def save_eig_red_hess(matrix): 
    """

    Parameters
    ----------
    matrix : np.ndarray
        The covariance matrix used for the reduced hessian decomposition

    Returns
    -------
    None.

    """
    redhessian=np.linalg.inv(matrix)
    np.savetxt('redhessian.out', redhessian, delimiter=',')
    eigH2, vH2 = np.linalg.eig(redhessian)
    np.savetxt('eigredhessian.out', eigH2)
    
    return None



"""
Reduced Hessian Generation

This module creates the reduced Hessian for use in various KIPET modules
"""
import os
from pathlib import Path
import time

import numpy as np
import pandas as pd
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.environ import (
    Constraint,
    Param,
    Set,
    SolverFactory,
    Suffix,
    )
from scipy.sparse import coo_matrix, triu
from scipy.sparse import csr_matrix, csc_matrix
from scipy.sparse.linalg import spsolve
    
from kipet.common.parameter_handling import (
    set_scaled_parameter_bounds,
    )
 
DEBUG = False

class ReducedHessian(object):
    """Class for handling the reduced hessian calculations in KIPET/NSD"""
    
    def __init__(self, 
                 model_object,
                 kkt_method = 'k_aug',
                 global_param_name = 'd',
                 param_con_method = 'global',
                 scaled = False,
                 rho = 10,
                 set_param_bounds = False,
                 parameter_set = None,
                 variable_name = 'P',
                 param_set_name = 'parameter_names',
                 current_set = 'current_p_set',
                 set_up_constraints = True,
                 use_duals = True,
                 global_constraint_name = 'fix_params_to_global',
                 file_number = None
                 ):
        
        self.model_object =  model_object
        self.kkt_method = kkt_method
        
        self.global_param_name = global_param_name
        self.global_constraint_name = global_constraint_name
        self.param_con_method = param_con_method
        self.scaled = scaled
        self.rho = rho
        self.set_param_bounds = set_param_bounds
        self.parameter_set = parameter_set
        self.variable_name = variable_name
        self.param_set_name = param_set_name
        self.current_set = current_set
        self.set_up_constraints = set_up_constraints
        self.use_duals = use_duals
        self.file_number = file_number
        
        self.verbose = DEBUG


    def get_tmp_file(self):
        
        file_tag = ''
        if self.file_number is not None:
            file_tag += f'_{self.file_number}'
            
        return 'ipopt_output' + file_tag

    def get_file_info(self):
        
        tmpfile_i = self.get_tmp_file()
            
        with open(tmpfile_i, 'r') as f:
            output_string = f.read()
        
        stub = output_string.split('\n')[0].split(',')[1][2:-4]
        
        nl_file = Path(stub + '.nl')
        col_file = Path(stub + '.col')
        row_file = Path(stub + '.row')
        sol_file = Path(stub + '.sol')
        
        self.sol_files = dict(
            nl = nl_file,
            col = col_file,
            row = row_file,
            sol = sol_file,
            )
        
        return None
    
    def delete_sol_files(self):
        
        if hasattr(self, 'sol_files'):
        
            for key, file in self.sol_files.items():
                file.unlink()
                
            del self.sol_files
            
        return None

    def get_kkt_info(self):
        
        """Takes the model and uses PyNumero to get the jacobian and Hessian
        information as dataframes
        
        Args:
            model_object (pyomo ConcreteModel): A pyomo model instance of the current
                problem (used in calculating the reduced Hessian)
    
            method (str): defaults to k_aug, method by which to obtain optimization
                results
    
        Returns:
            
            kkt_data (dict): dictionary with the following structure:
                
                    {
                    'J': J,   # Jacobian
                    'H': H,   # Hessian
                    'var_ind': var_index_names, # Variable index
                    'con_ind': con_index_names, # Constraint index
                    'duals': duals, # Duals
                    }
            
        """
        self.get_file_info()
        
        if self.kkt_method == 'pynumero':
        
            nlp = PyomoNLP(self.model_object)
            varList = nlp.get_pyomo_variables()
            conList = nlp.get_pyomo_constraints()
            duals = nlp.get_duals()
            
            J = nlp.extract_submatrix_jacobian(pyomo_variables=varList, pyomo_constraints=conList)
            H = nlp.extract_submatrix_hessian_lag(pyomo_variables_rows=varList, pyomo_variables_cols=varList)
            J = csc_matrix(J)
            
            var_index_names = [v.name for v in varList]
            con_index_names = [v.name for v in conList]
            
        elif self.kkt_method == 'k_aug':
        
            kaug = SolverFactory('k_aug')
            
            kaug.options["deb_kkt"] = ""  
            kaug.solve(self.model_object, tee=False)
            
            hess = pd.read_csv('hess_debug.in', delim_whitespace=True, header=None, skipinitialspace=True)
            hess.columns = ['irow', 'jcol', 'vals']
            hess.irow -= 1
            hess.jcol -= 1
            os.unlink('hess_debug.in')
            
            jac = pd.read_csv('jacobi_debug.in', delim_whitespace=True, header=None, skipinitialspace=True)
            m = jac.iloc[0,0]
            n = jac.iloc[0,1]
            jac.drop(index=[0], inplace=True)
            jac.columns = ['irow', 'jcol', 'vals']
            jac.irow -= 1
            jac.jcol -= 1
            os.unlink('jacobi_debug.in')
            
            #try:
            #    duals = read_duals(stub + '.sol')
            #except:
            duals = None
            
            J = coo_matrix((jac.vals, (jac.irow, jac.jcol)), shape=(m, n)) 
            Hess_coo = coo_matrix((hess.vals, (hess.irow, hess.jcol)), shape=(n, n)) 
            H = Hess_coo + triu(Hess_coo, 1).T
            
            var_index_names = pd.read_csv(self.sol_files['col'], sep = ';', header=None) # dummy sep
            con_index_names = pd.read_csv(self.sol_files['row'], sep = ';', header=None) # dummy sep
            
            var_index_names = [var_name for var_name in var_index_names[0]]
            con_index_names = [con_name for con_name in con_index_names[0].iloc[:-1]]
            con_index_number = {v: k for k, v in enumerate(con_index_names)}
        
        self.delete_sol_files()
        
        self.kkt_data = {
                    'J': J,
                    'H': H,
                    'var_ind': var_index_names,
                    'con_ind': con_index_names,
                    'duals': duals,
                    }
        
        return None
    
    def get_kkt_df(self):
        
        self.get_kkt_info()
        
        H = self.kkt_data['H']
        var_index_names = self.kkt_data['var_ind']
        con_index_names = self.kkt_data['con_ind']
        
        H_df = pd.DataFrame(H.todense(), columns=var_index_names, index=var_index_names)
    
        # See if this works
    
        # h_con_indx = [k for k in self.model.C.keys()]
        # h_con = [f'Z[{h[0]},{h[1]}]' for h in h_con_indx if h[1] in self.model.measured_data]
        # h_con_indx = [k for k in self.model.U.keys()]
        # h_con += [f'X[{h[0]},{h[1]}]' for h in h_con_indx  if h[1] in self.model.measured_data]
        
        
        # col_ind  = [var_ind.loc[var_ind[0] == v].index[0] for v in h_con]
        # Zr = Z[col_ind, :]
        
        # df_Zr = pd.DataFrame(Zr, index=h_con, columns=[k for k, v in self.model.P.items()])
        # df_Hv = H_df.loc[h_con, h_con]    
    
    
        return H_df
    
    
    
    def prep_model_for_k_aug(self):
        """This function prepares the optimization models with required
        suffixes.
        
        Args:
            model_object (pyomo model): The model of the system
            
        Retuns:
            None
            
        """
        self.model_object.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
        # self.model_object.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
        # self.model_object.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
        # self.model_object.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
        # self.model_object.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)
        # self.model_object.red_hessian = Suffix(direction=Suffix.EXPORT)
        # self.model_object.dof_v = Suffix(direction=Suffix.EXPORT)
        # self.model_object.rh_name = Suffix(direction=Suffix.IMPORT)
        
        count_vars = 1
        for k, v in self.model_object.P.items():
            self.model_object.dof_v[k] = count_vars
            count_vars += 1
        
        self.model_object.npdp = Suffix(direction=Suffix.EXPORT)
        
        return None
    
    def calculate_duals(self):
        """Get duals"""
    
        # For testing - very slow and should not be used!
        if self.kkt_method == 'pynumero':
        
            nlp = PyomoNLP(self.model_object)
            varList = nlp.get_pyomo_variables()
            conList = nlp.get_pyomo_constraints()
            duals = nlp.get_duals()
            
            J = nlp.extract_submatrix_jacobian(pyomo_variables=varList, pyomo_constraints=conList)
            H = nlp.extract_submatrix_hessian_lag(pyomo_variables_rows=varList, pyomo_variables_cols=varList)
            J = csc_matrix(J)
            
            var_index_names = [v.name for v in varList]
            con_index_names = [v.name for v in conList]                       
            
            dummy_constraints = [f'{self.global_constraint_name}[{k}]' for k in self.parameter_set]
            jac_row_ind = [con_index_names.index(d) for d in dummy_constraints] 
            duals_imp = [duals[i] for i in jac_row_ind]
        
            self.duals = dict(zip(self.parameter_set, duals_imp))
            if self.verbose:
                print(f'The pynumero results are:')
                print(self.duals)
            
        else:
                         
            self.duals = {key: self.model_object.dual[getattr(self.model_object, self.global_constraint_name)[key]] for key, val in getattr(self.model_object, self.global_param_name).items()}
        
            if self.verbose:
                print('The duals are:')
                print(self.duals)
        
        self.delete_sol_files()
        
        return self.duals              
    
    def optimize_model(self, d=None):
        """Takes the model object and performs the optimization
        
        Args:
            model_object (pyomo model): the pyomo model of the reaction
            
            parameter_set (list): list of current model parameters
            
        Returns:
            reduced_hessian (numpy array): reduced hessian of the model
        
        """
        if self.verbose:    
            print(f'd: {d}')
        
        ipopt = SolverFactory('ipopt')
        tmpfile_i = self.get_tmp_file()
    
        if self.param_con_method == 'global':
            
            if d is not None:
                param_val_dict = {p: d[i] for i, p in enumerate(self.parameter_set)}
                for k, v in getattr(self.model_object, self.variable_name).items():
                    v.set_value(param_val_dict[k])
        
            self.add_global_constraints() 
        
        elif self.param_con_method == 'fixed':
            
            if hasattr(self.model_object, self.global_constraint_name):
                self.model_object.del_component(self.global_constraint_name)  
        
            delta = 1e-20  
            for k, v in getattr(self.model_object, self.variable_name).items():
                if k in self.parameter_set:
                    ub = v.value
                    lb = v.value - delta
                    v.setlb(lb)
                    v.setub(ub)
                    v.unfix()
                else:
                    v.fix()
                    
        if self.set_param_bounds:
            set_scaled_parameter_bounds(self.model_object, 
                                        parameter_set=self.parameter_set, 
                                        rho=self.rho, 
                                        scaled=self.scaled)  
                    
        ipopt.solve(self.model_object, 
                    symbolic_solver_labels=True, 
                    keepfiles=True, 
                    tee=False,
                    logfile=tmpfile_i,
                    )
        
        # Create the file object so that it can be deleted
        self.get_file_info()
        
        return None

    def calculate_reduced_hessian(self, d=None, optimize=False, return_Z=False):
        """Calculate the reduced Hessian
        
        Args:
            model_object (pyomo model): the pyomo model of the reaction
            
            parameter_set (list): list of current model parameters
            
        Returns:
            reduced_hessian (numpy array): reduced hessian of the model
        
        """
        if optimize:
            self.optimize_model(d)
        
        self.get_kkt_info()
        H = self.kkt_data['H']
        J = self.kkt_data['J']
        var_ind = self.kkt_data['var_ind']
        con_ind_new = self.kkt_data['con_ind']
        duals = self.kkt_data['duals']
    
        col_ind = [var_ind.index(f'{self.variable_name}[{v}]') for v in self.parameter_set]
        m, n = J.shape  
     
        if self.param_con_method == 'global':
            
            dummy_constraints = [f'{self.global_constraint_name}[{k}]' for k in self.parameter_set]
            jac_row_ind = [con_ind_new.index(d) for d in dummy_constraints] 
            #duals_imp = [duals[i] for i in jac_row_ind]
            
            #print(J.shape, len(duals_imp))
    
            J_c = delete_from_csr(J.tocsr(), row_indices=jac_row_ind).tocsc()
            row_indexer = SparseRowIndexer(J_c)
            J_f = row_indexer[col_ind]
            J_f = delete_from_csr(J_f.tocsr(), row_indices=jac_row_ind, col_indices=[])
            J_l = delete_from_csr(J_c.tocsr(), col_indices=col_ind)  
    
        elif self.param_con_method == 'fixed':
            
            jac_row_ind = []
            duals_imp = None
        
            J_c = J.tocsc()# delete_from_csr(J.tocsr(), row_indices=jac_row_ind).tocsc()
            row_indexer = SparseRowIndexer(J_c.T)
            J_f = row_indexer[col_ind].T
            #J_f = delete_from_csr(J_f.tocsr(), row_indices=jac_row_ind, col_indices=[]) 
            J_l = delete_from_csr(J_c.tocsr(), col_indices=col_ind)  
            
        else:
            None
        
        r_hess, Z_mat = self._reduced_hessian_matrix(J_f, J_l, H, col_ind)
       
        if not return_Z:
            return r_hess.todense()
        else:
            return r_hess.todense(), Z_mat

    @staticmethod
    def _reduced_hessian_matrix(F, L, H, col_ind):
        """This calculates the reduced hessian by calculating the null-space based
        on the constraints
        
        Args:
            F (csr_matrix): Rows of the Jacobian related to fixed parameters
            
            L (csr_matrix): The remainder of the Jacobian without parameters
            
            H (csr_matrix): The sparse Hessian
            
            col_ind (list): indicies of columns with fixed parameters
        
        Returns:
            reduced_hessian (csr_matrix): sparse version of the reduced Hessian
            
        """
        n = H.shape[0]
        n_free = n - F.shape[0]
        
        X = spsolve(L.tocsc(), -F.tocsc())
        
        col_ind_left = list(set(range(n)).difference(set(col_ind)))
        col_ind_left.sort()
        
        Z = np.zeros([n, n_free])
        Z[col_ind, :] = np.eye(n_free)
        
        if isinstance(X, csc_matrix):
            Z[col_ind_left, :] = X.todense()
        else:
            Z[col_ind_left, :] = X.reshape(-1, 1)
            
        Z_mat = coo_matrix(np.mat(Z)).tocsr()
        Z_mat_T = coo_matrix(np.mat(Z).T).tocsr()
        Hess = H.tocsr()
        reduced_hessian = Z_mat_T * Hess * Z_mat
        
        return reduced_hessian, Z_mat

    def add_global_constraints(self):
        """This adds the dummy constraints to the model forcing the local
        parameters to equal the current global parameter values
         
        Args:
            model_object (pyomo ConcreteModel): Pyomo model to add constraints to
            
            parameter_set (list): List of parameters to fix using constraints
            
            scaled (bool): True if scaled, False if not scaled
             
        Returns:
            None
         
        """
        if self.parameter_set is None:
            self.parameter_set = [p for p in getattr(self.model_object, self.variable_name)]
         
        if self.scaled:
            global_param_init = {p: 1 for p in self.parameter_set}
        else:
            global_param_init = {p: getattr(self.model_object, self.variable_name)[p].value for p in self.parameter_set}
    
        for comp in [self.global_constraint_name, self.global_param_name, self.current_set]:
            if hasattr(self.model_object, comp):
                self.model_object.del_component(comp)   
         
        setattr(self.model_object, self.current_set, Set(initialize=self.parameter_set))
    
        setattr(self.model_object, self.global_param_name, 
                Param(getattr(self.model_object, self.param_set_name),
                              initialize=global_param_init,
                              mutable=True,
                              ))
         
        def rule_fix_global_parameters(m, k):
            return getattr(m, self.variable_name)[k] - getattr(m, self.global_param_name)[k] == 0
             
        setattr(self.model_object, self.global_constraint_name, 
                Constraint(getattr(self.model_object, self.current_set),
                           rule=rule_fix_global_parameters)
            )
        
        return None
 
def read_duals(sol_file):
    """Reads the duals from the sol file after solving the problem
    
    Args:
        sol_file (str): The absolute path to the sol file
        
    Returns:
        duals (list): The list of duals values taken from the sol file
    
    """
    sol_file_abs = Path(sol_file)
    
    duals = []
    with sol_file_abs.open() as f: 
        lines = f.readlines()
       
    lines_found = True
    num_of_vars = int(lines[9])
      
    for ln, line in enumerate(lines):
        line = line.rstrip('\n')
        line = line.lstrip('\t').lstrip(' ')
        
        if ln >= 12 and ln <= (num_of_vars + 11):
            duals.append(float(line))
            
    return duals

def delete_from_csr(mat, row_indices=[], col_indices=[]):
    """
    Remove the rows (denoted by ``row_indices``) and columns (denoted by 
    ``col_indices``) from the CSR sparse matrix ``mat``.
    WARNING: Indices of altered axes are reset in the returned matrix
    
    Args:
        mat (csr_matrix): Sparse matrix to delete rows and cols from
        
        row_indicies (list): rows to delete
        
        col_indicies (list): cols to delete
        
    Returns:
        mat (csr_matrix): The sparse matrix with the rows and cols removed
    
    """
    if not isinstance(mat, csr_matrix):
        raise ValueError("works only for CSR format -- use .tocsr() first")

    rows = []
    cols = []
    if row_indices:
        rows = list(row_indices)
    if col_indices:
        cols = list(col_indices)

    if len(rows) > 0 and len(cols) > 0:
        row_mask = np.ones(mat.shape[0], dtype=bool)
        row_mask[rows] = False
        col_mask = np.ones(mat.shape[1], dtype=bool)
        col_mask[cols] = False
        return mat[row_mask][:,col_mask]
    elif len(rows) > 0:
        mask = np.ones(mat.shape[0], dtype=bool)
        mask[rows] = False
        return mat[mask]
    elif len(cols) > 0:
        mask = np.ones(mat.shape[1], dtype=bool)
        mask[cols] = False
        return mat[:,mask]
    else:
        return mat

class SparseRowIndexer:
    """Class used to splice sparse matrices"""
    
    def __init__(self, matrix):
        data = []
        indices = []
        indptr = []
        
        _class = 'csr'
        #print(f'LOOK HERE: {type(matrix)}')
        if isinstance(matrix, csc_matrix):
            _class = 'csc'
       #     print(_class)
        
        self._class = _class
        # Iterating over the rows this way is significantly more efficient
        # than matrix[row_index,:] and matrix.getrow(row_index)
        for row_start, row_end in zip(matrix.indptr[:-1], matrix.indptr[1:]):
             data.append(matrix.data[row_start:row_end])
             indices.append(matrix.indices[row_start:row_end])
             indptr.append(row_end-row_start) # nnz of the row

        self.data = np.array(data)
        self.indices = np.array(indices)
        self.indptr = np.array(indptr)
        self.n_columns = matrix.shape[1]

    def __getitem__(self, row_selector):
        data = np.concatenate(self.data[row_selector])
        indices = np.concatenate(self.indices[row_selector])
        indptr = np.append(0, np.cumsum(self.indptr[row_selector]))

        shape = [indptr.shape[0]-1, self.n_columns]
        
        if self._class == 'csr':
            return csr_matrix((data, indices, indptr), shape=shape)
        else:
            return csr_matrix((data, indices, indptr), shape=shape).T.tocsc()


"""TemplateBuilder - handles inputs for generating the Pyomo model"""

# Standard library imports
import copy
import inspect
import itertools
import logging
import numbers
import six
import sys
import warnings

# Third party imports
import numpy as np
import pandas as pd
from pyomo.environ import *
from pyomo.dae import *
from pyomo.environ import units as u

# KIPET library imports
from kipet.post_model_build.scaling import scale_parameters
from kipet.core_methods.PyomoSimulator import PyomoSimulator
from kipet.top_level.variable_names import VariableNames
from kipet.post_model_build.pyomo_model_tools import get_index_sets
from kipet.common.VisitorClasses import ReplacementVisitor
from kipet.common.component_expression import Comp

logger = logging.getLogger('ModelBuilderLogger')


class TemplateBuilder(object):
    """Helper class for creation of models.

    Attributes:
        _component_names (set): container with names of components.

        _parameters (dict): map of parameter names to corresponding values

        _parameters_bounds (dict): map of parameter names to bounds tuple

        _init_conditions (dict): map of component/state name to its initial condition

        _spectal_data (DataFrame, optional): DataFrame with time indices and wavelength columns

        _absorption_data (DataFrame, optional): DataFrame with wavelength indices and component names columns

        _odes (Function): function specified by user to return dictionary with ODE expressions

        _meas_times (set, optional): container of measurement times

        _huplcmeas_times (set, optional): container of  H/UPLC measurement times

        _allmeas_times (set, optional): container of all measurement times

        _feed_times (set, optional): container of feed times

        _complementary_states (set,optional): container with additional states

    """
    __var = VariableNames()

    def __init__(self, **kwargs):
        """Template builder constructor

        Args:
            **kwargs: Arbitrary keyword arguments.
            concentrations (dictionary): map of component name to initial condition

            parameters (dictionary): map of parameter name to its corresponding value

            extra_states (dictionary): map of state name to initial condition

        """
        self._smoothparameters = dict()  # added for mutable parameters CS
        self._smoothparameters_mutable = dict() #added for mutable parameters CS

        #added for initial condition parameter estimates: CS
        self._initextraparams = dict()
        self._initextraparams_init = dict()  # added for parameter initial guess CS
        self._initextraparams_bounds = dict()

        # TODO: Put all data into it's own structure (attrs?)

        self._y_bounds = dict()  # added for additional optional bounds CS
        self._y_init = dict()
        self._prof_bounds = list()  # added for additional optional bounds MS
       # self._init_conditions = dict()
        self._spectral_data = None
        self._concentration_data = None
        self._complementary_states_data = None # added for complementary state data (Est.) KM
        self._custom_data = None
        self._custom_objective = None
        self._huplc_data = None #added for additional data CS
        self._smoothparam_data = None  # added for additional smoothing parameter data CS
        self._absorption_data = None
        self._odes = None
        self._meas_times = set()
        self._huplcmeas_times = set() #added for additional data CS
        self._allmeas_times = set() #added for new data structure CS
        self._m_lambdas = None
        self._complementary_states = set()
        self._algebraics = dict()
        self._algebraic_constraints = None
        self._known_absorbance = None
        self._is_known_abs_set = False
        self._warmstart = False  # add for warmstart CS
        self._known_absorbance_data = None
        self._non_absorbing = None
        self._is_non_abs_set = False
        self._huplc_absorbing = None #for add additional data CS
        self._is_huplc_abs_set = False #for add additional data CS
        self._vol = None #For add additional data CS
        self._feed_times = set()  # For inclusion of discrete feeds CS
        self._is_D_deriv = False
        self._is_C_deriv = False
        self._is_U_deriv = False
        self._is_Dhat_deriv = False
        self._state_sigmas = None # Need to put sigmas into the pyomo model as params
        self._model_constants = None # Used in EstimaationPotential
        self._scale_parameters = False # Should be True for EstimationPotential (automatic)
        self._times = None
        self._all_state_data = list()
        
        # bounds and init for unwanted contributions KH.L
        self._G_contribution = None
        
        self._qr_bounds = None
        self._qr_init = None
        self._g_bounds = None
        self._g_init = None

        self._add_dosing_var = False

        
    def add_state_variance(self, sigma_dict):
        """Provide a variance for the measured states
        
        Args:
            sigma (dict): A dictionary of the measured states with known 
            variance. Provide the value of sigma (standard deviation).
            
        Returns:
            None
        """
        self._state_sigmas = sigma_dict 
        
        return None

    def set_parameter_scaling(self, use_scaling: bool):
        """Makes an option to use the scaling method implemented for estimability
        
        Args:
            use_scaling (bool): Defaults to False, for using scaled parameters
            
        Returns:
            None
        
        """
        self._scale_parameters = use_scaling
    
        return None
        
    def set_model_times(self, times):
        """Add the times to the builder template.
        
        Args:
            times (iterable): start and end times for the pyomo model
        
        Returns:
            None
            
        """
        self._times = times
        
    """
    Adding in the model elements
    """  
    def add_model_element(self, BlockObject):
        
        data_type = BlockObject.attr_class_set_name.rstrip('s')
        block_attr_name = f'template_{data_type}_data'
        setattr(self, block_attr_name, BlockObject)
        return None
 
    def add_smoothparameter(self, *args, **kwds):
        """Add a kinetic parameter(s) to the model.

        Note:
            Plan to change this method add parameters as PYOMO variables

            This method tries to mimic a template implementation. Depending
            on the argument type it will behave differently

        Args:
            param1 (boolean): Mutable value. Creates a variable mutable

        Returns:
            None

        """
        mutable = kwds.pop('mutable', False)

        if len(args) == 2:
            first = args[0]
            second = args[1]
            if isinstance(first, six.string_types):
                self._smoothparameters[first] = second
                if mutable is not False and second is pd.DataFrame:
                    self._smoothparameters_mutable[first] = mutable #added for mutable parameters CS
            else:
                raise RuntimeError('Parameter argument not supported. Try pandas.Dataframe and mutable=True')
        else:
            raise RuntimeError('Parameter argument not supported. Try pandas.Dataframe and mutable=True')
    
    def input_data(self, data_block_dict=None, spectral_data=None):
        """This should take a DataContainer and input the data, but for now
        it will take a DataBlock list
        
        """
        time_span = 0
        time_conversion_factor = 1
        
        data_type_labels = {'component' : 'concentration',
                            'state': 'complementary_states',
                            'algebraic': 'custom',
                            }

        if data_block_dict is not None:
            for data_type, data_label in data_type_labels.items():
                if hasattr(self, f'template_{data_type}_data'):
                    c_info = getattr(self, f'template_{data_type}_data')
                    for comp in c_info:
                        if hasattr(comp, 'data_link'):
                            data_block = data_block_dict[comp.data_link]
                            time_span = max(time_span, data_block.time_span[1]*time_conversion_factor)
                            data_frame = data_block.data[comp.name]*comp.conversion_factor
                            # data_frame.index = data_frame.index*time_conversion_factor
                            self._add_state_data(data_frame, data_label, overwrite=False)

        # Spectral data is handled differently
        if spectral_data is not None:
            self._add_state_data(spectral_data.data, 'spectral')
            # spectral_data.data.index = spectral_data.data.index*time_conversion_factor
            time_span = max(time_span, spectral_data.data.index.max())
                
        self.time_span_max = time_span
       
        return None
    
    def clear_data(self):
        
        self._spectral_data = None
        self._concentration_data = None
        self._complementary_states_data = None # added for complementary state data (Est.) KM
        self._huplc_data = None #added for additional data CS
        self._smoothparam_data = None  # added for additional smoothing parameter data CS
        self._absorption_data = None
        
        return None

    def _add_state_data(self, data, data_type, label=None, overwrite=True):
        """Generic method for adding data (concentration or complementary 
        state data) - uses the measured data attribute to process
        
        Args:
            data (DataFrame): DataFrame with measurement times as
                              indices and concentrations as columns.
                              
            data_type (str): The name of the attribute for where the data is to
                be stored.
                
            label (str): The label used to descibe the data in the pyomo model
                index.

        Returns:
            None

        """
        built_in_data_types = {
            'concentration' : self.__var.concentration_measured,
            'complementary_states' : self.__var.state,
            'spectral' : self.__var.spectra_data,
            'huplc' : self.__var.huplc_data,
            'smoothparam' : self.__var.smooth_parameter,
            'custom' : self.__var.user_defined,
            }
        
        state_data = [
            self.__var.concentration_spectra,
            self.__var.state,
            self.__var.concentration_measured,
            self.__var.user_defined,
            ]
            
        deriv_data = [
            self.__var.concentration_spectra,
            self.__var.state,
            self.__var.spectra_data,
            self.__var.huplc_data,
            self.__var.concentration_measured,
            self.__var.user_defined,
            ]
        
        if label is None:
            try:
                label = built_in_data_types[data_type]
            except:
                raise ValueError("You need to provide a label for custom data types")
        
        if isinstance(data, pd.Series):
            data = pd.DataFrame(data)
        
        if isinstance(data, pd.DataFrame):
            dfc = pd.DataFrame(index=self._feed_times, columns=data.columns)
            for t in self._feed_times:
                if t not in data.index:
                    dfc.loc[t] = [0.0 for n in range(len(data.columns))]
      
            dfallc = data.append(dfc)
            dfallc.sort_index(inplace=True)
            dfallc.index = dfallc.index.to_series().apply(
                lambda x: np.round(x, 6))

            count = 0
            for j in dfallc.index:
                if count >= 1 and count < len(dfallc.index):
                    if dfallc.index[count] == dfallc.index[count - 1]:
                        dfallc = dfallc.dropna()
                        
                count += 1

            dfallc = dfallc.dropna(how='all')

            if not overwrite:
                if hasattr(self, f'_{data_type}_data'):
                    df_data = getattr(self, f'_{data_type}_data')
                    #print(f'df_data:\n{df_data}')
                    df_data = pd.concat([df_data, dfallc], axis=1)
                    setattr(self, f'_{data_type}_data', df_data)
            else:
                setattr(self, f'_{data_type}_data', dfallc)
            
            if label in state_data:
                self._all_state_data += list(data.columns)
        else:
            raise RuntimeError(f'{data_type.capitalize} data format not supported. Try pandas.DataFrame')
        
        if label in deriv_data:
            C = np.array(dfallc)
            setattr(self, f'_is_{label}_deriv', False)
            for t in range(len(dfallc.index)):
                for l in range(len(dfallc.columns)):
                    if C[t, l] >= 0 or np.isnan(C[t, l]):
                        pass
                    else:
                        setattr(self, f'_is_{label}_deriv', True)
            #print(label, hasattr(self, f'_is_{label}_deriv'))
            if getattr(self, f'_is_{label}_deriv') == True:
                print(
                    f"Warning! Since {label}-matrix contains negative values Kipet is assuming a derivative of {label} has been inputted")

        return None
           
    def add_huplc_data(self, data, overwrite=True): #added for the inclusion of h/uplc data CS
        """Add HPLC or UPLC data as a wrapper to _add_state_data

                Args:
                    data (DataFrame): DataFrame with measurement times as
                                      indices and wavelengths as columns.

                Returns:
                    None
        """
        self._add_state_data(data,
                             data_type='huplc',
                             overwrite=overwrite)

    def add_smoothparam_data(self, data, overwrite=True): #added for mutable smoothing parameter option CS
        """Add smoothing parameters as a wrapper to _add_state_data

                Args:
                    data (DataFrame): DataFrame with measurement times as
                                      indices and wavelengths as columns.

                Returns:
                    None
        """
        self._add_state_data(data,
                             data_type='smoothparam',
                             overwrite=overwrite)


    def add_absorption_data(self, data, overwrite=True):
        """Add absorption data

        Args:
            data (DataFrame): DataFrame with wavelengths as
                              indices and muxture components as columns.

        Returns:
            None

        """
        if isinstance(data, pd.DataFrame):
            self._absorption_data = data
        else:
            raise RuntimeError('Spectral data format not supported. Try pandas.DataFrame')

    def round_time(self, time):
        
        return round(time, 6)

    # For inclusion of discrete jumps
    def add_feed_times(self, times):
        """Add measurement times to the model

        Args:
            times (array_like): feeding points

        Returns:
            None

        """
        for t in times:
            t = self.round_time(t)  # for added ones when generating data otherwise too many digits due to different data types CS
            self._feed_times.add(t)
            self._meas_times.add(t)  # added here to avoid double addition CS

    # For inclusion of discrete jumps
    def add_measurement_times(self, times):
        """Add measurement times to the model

        Args:
            times (array_like): measurement points

        Returns:
            None

        """
        for t in times:
            t = self.round_time(t)  # for added ones when generating data otherwise too many digits due to different data types CS
            self._meas_times.add(t)

    # Why is this function here?
    def add_huplcmeasurement_times(self, times): #added for additional huplc times that are on a different time scale CS
        """Add H/UPLC measurement times to the model

        Args:
            times (array_like): measurement points

        Returns:
            None

        """
        for t in times:
            t = self.round_time(t)  # for added ones when generating data otherwise too many digits due to different data types CS
            self._huplcmeas_times.add(t)

    # read bounds and initialize for qr and g (unwanted contri variables) from users KH.L
    def add_qr_bounds_init(self, **kwds):
        bounds = kwds.pop('bounds', None)
        init = kwds.pop('init', None)
        
        self._qr_bounds = bounds
        self._qr_init = init
        
        return None
        
    def add_g_bounds_init(self, **kwds):
        bounds = kwds.pop('bounds', None)
        init = kwds.pop('init', None)
        
        self._g_bounds = bounds
        self._g_init = init
    
        return None
    
    def add_dosing_var(self, n_steps):
        
        self._add_dosing_var = True
        self._number_of_steps = n_steps
    
    def set_odes_rule(self, rule):
        """Set the ode expressions.

        Defines the ordinary differential equations that define the dynamics of the model

        Args:
            rule (dict) : Model expressions for the rate equations

        Returns:
            None

        """
        self._odes = rule
        
        return None

    def set_algebraics_rule(self, rule, asdict=False):
        """Set the algebraic expressions.

        Defines the algebraic equations for the system

        Args:
            rule (function): Python function that returns a list of tuples

        Returns:
            None

        """
        if not asdict:
            inspector = inspect.getargspec(rule)
            if len(inspector.args) != 2:
                raise RuntimeError('The rule should have two inputs')
            self._algebraic_constraints = rule
        else:
            self._algebraic_constraints = rule
            self._use_alg_dict = True
        
        return None

    def set_objective_rule(self, algebraic_vars):
        """Set the algebraic expressions.

        Defines the algebraic equations for the system

        Args:
            algebraic_vars (str, list): str or list of algebraic variables

        Returns:
            None

        """
        if not isinstance(algebraic_vars, list):
            algebraic_vars = list(algebraic_vars)
        self._custom_objective = algebraic_vars
        
        return None

    def bound_profile(self, var, bounds, comp=None, profile_range=None):
        """function that allows the user to bound a certain profile to some value

        Args:
            var (pyomo variable object): the pyomo variable that we will bound
            comp (str, optional): The component that bound applies to
            profile_range (tuple,optional): the range within the set to be bounded
            bounds (tuple): the values to bound the profile to

        Returns:
            None

        """
        if not isinstance(var, str):
            raise RuntimeError('var argument needs to be type string')

        if var not in ['C', 'U', 'S']:
            raise RuntimeError('var argument needs to be either C, U, or S')

        if comp is not None:
            if not isinstance(comp, str):
                raise RuntimeError('comp argument needs to be type string')
            if comp not in self._component_names:
                raise RuntimeError('comp needs to be one of the components')

        if profile_range is not None:
            if not isinstance(profile_range, tuple):
                raise RuntimeError('profile_range needs to be a tuple')
                if profile_range[0] > profile_range[1]:
                    raise RuntimeError('profile_range[0] must be greater than profile_range[1]')

        if not isinstance(bounds, tuple):
            raise RuntimeError('bounds needs to be a tuple')

        self._prof_bounds.append([var, comp, profile_range, bounds])

    def _validate_data(self, model):
        """Verify all inputs to the model make sense.

        This method is not suppose to be used by users. Only for developers use

        Args:
            model (pyomo or casadi model): Model

            start_time (float): initial time considered in the model

            end_time (float): final time considered in the model

        Returns:
            None

        """
        if not self.template_component_data.component_set('concentration'):
            warnings.warn('The Model does not have any mixture components')
        else:
            if self._odes:
                if isinstance(self._odes, dict):
                    dummy_balances = self._odes
                else:
                    dummy_balances = self._odes(model, model.start_time)
                if len(self._component_names) + len(self._complementary_states) != len(dummy_balances):
                    print(
                        'WARNING: The number of ODEs is not the same as the number of state variables.\n If this is the desired behavior, some odes must be added after the model is created.')

            else:
                print(
                    'WARNING: differential expressions not specified. Must be specified by user after creating the model')

        if self._algebraics:
            if self._algebraic_constraints and not isinstance(self._algebraic_constraints, dict):
                dummy_balances = self._algebraic_constraints(model, model.start_time)
                if len(self._algebraics) != len(dummy_balances):
                    print(
                        'WARNING: The number of algebraic equations is not the same as the number of algebraic variables.\n If this is the desired behavior, some algebraics must be added after the model is created.')
            else:
                print(
                    'WARNING: algebraic expressions not specified. Must be specified by user after creating the model')

        if self._absorption_data is not None:
            if not self._meas_times:
                raise RuntimeError('Need to add measurement times')


    def _add_algebraic_var(self, model):
        """If algebraics are present, add the algebraic variable to the model
        """
        if hasattr(self, 'template_algebraic_data'):    
            a_info = self.template_algebraic_data
        
            setattr(model, self.__var.algebraic, Var(model.alltime,
                                                     a_info.names,
                                                     initialize=a_info.as_dict('value')))
                                                     
            for alg in a_info:
                if alg.bounds is not None:
                    for t in model.alltime:
                        getattr(model, self.__var.algebraic)[t, alg.name].setlb(alg.lb)
                        getattr(model, self.__var.algebraic)[t, alg.name].setub(alg.ub)
                        
            # for alg in a_info:
            #     if alg.step is not None:
                     
            #     setattr(model, self.__var.step_variable, Var(model.alltime,
            #                                                  steps,
            #                                                  initialize=1.0))
                
            #     #Constraint to keep track of the step values (for diagnostics)
            #     def rule_step_function(m, t, k):
            #         step_const = getattr(m, self.__var.step_variable)[t, k] - step_fun(m, t, num=k, **self._step_data[k])
            #         return step_const == 0.0
    
            #     constraint_suffix = 'constraint'
                
            #     setattr(model, self.__var.step_variable + constraint_suffix,
            #             Constraint(model.alltime,
            #                        steps,
            #                        rule=rule_step_function))
    
        return None
    
    def _add_initial_conditions(self, model):
        """Set up the initial conditions for the model"""
        
        var_info = dict(self.template_component_data._dict)
        
        if hasattr(self, 'template_state_data'):
            var_info.update(self.template_state_data._dict)
        
        model.init_conditions = Var(model.states,
                                    initialize={k: v.value for k, v in var_info.items()},
                                    )
        unknown_init = {}
        
        for var, obj in model.init_conditions.items():
            if var_info[var].known:
                obj.fix()
            else:
                lb = var_info[var].lb
                ub = var_info[var].ub
                init = var_info[var].value
                obj.setlb(lb)
                obj.setub(ub)
                unknown_init[var] = init
        
        if len(unknown_init) > 0:
            model._unknown_init_set = Set(initialize=list(unknown_init.keys()))
            model.Pinit = Var(model._unknown_init_set,
                              initialize=unknown_init)
            
            model.del_component('P_all')    
            model.P_all = Set(initialize=model.parameter_names | model._unknown_init_set,
                                 ordered=True)

        def rule_init_conditions(m, k):
            if k in m.mixture_components:
                return getattr(m, self.__var.concentration_model)[m.start_time, k] - m.init_conditions[k] == 0
            else:
                return getattr(m, self.__var.state_model)[m.start_time, k] - m.init_conditions[k] == 0

        model.init_conditions_c = \
            Constraint(model.states, rule=rule_init_conditions)
            
        if hasattr(model, 'Pinit'):
            
            def rule_Pinit_conditions(m, k):
                if k in m.mixture_components:
                    return m.Pinit[k] - m.init_conditions[k] == 0
                else:
                    return m.Pinit[k] - m.init_conditions[k] == 0
    
            model.Pinit_conditions_c = \
                Constraint(model._unknown_init_set, rule=rule_Pinit_conditions)

        return None

    def _add_model_variables(self, model):
        """Adds the model variables to the pyomo model"""
        
        c_info = self.template_component_data
        self._component_names = c_info.names
        
        var_set = [c_info]
        
        if hasattr(self, 'template_state_data'):
            s_info = self.template_state_data
            self._complementary_states = s_info.names
            var_set.append(s_info)
        
        for var_class in var_set:
        
            v_info = var_class    
        
            model_pred_var_name = {
                    'components' : [self.__var.concentration_model, model.mixture_components],
                    'states' : [self.__var.state_model, model.complementary_states],
                        }
            
            var, model_set = model_pred_var_name[var_class.attr_class_set_name]
            
            # Check if any data for the variable type exists (Pyomo 5.7 update)
            if hasattr(model_set, 'ordered_data') and len(model_set.ordered_data()) == 0:
                continue
            
            setattr(model, var, Var(model.alltime,
                                          model_set,
                                          # units=v_info.as_dict('units'),
                                          initialize=1) 
                    )    
        
            for time, comp in getattr(model, var):
                if time == model.start_time.value:
                    
                    getattr(model, var)[time, comp].value = v_info[comp].value
                   
            setattr(model, f'd{var}dt', DerivativeVar(getattr(model, var),
                                                     # units=self.u.mol/self.u.l,
                                                      wrt=model.alltime)
                    )

        # Variables of provided data - set as fixed variables complementary to above
        fixed_var_name = {
                self.__var.concentration_measured : self._concentration_data,
                self.__var.state : self._complementary_states_data,
                self.__var.user_defined : self._custom_data,
                    }
        
        for var, data in fixed_var_name.items():
           
            if data is None or len(data) == 0:
                continue
            
            c_dict = dict()
            if hasattr(self, f'_is_{var}_deriv'):
                if getattr(self, f'_is_{var}_deriv') == True:
                    c_bounds = (None, None)
                else:
                    c_bounds = (0.0, None)
                
                if data is not None:
                    
                    for i, row in data.iterrows():
                         c_dict.update({(i, col): float(row[col]) for col in data.columns if not np.isnan(float(row[col]))})
                
                    setattr(model, f'{var}_indx', Set(initialize=list(c_dict.keys()), ordered=True))
                    
                    setattr(model, var, Var(getattr(model, f'{var}_indx'),
                                                  bounds=c_bounds,
                                                  initialize=c_dict,
                                                  )
                            )
                    
                    for k, v in getattr(model, var).items():
                        getattr(model, var)[k].fixed = True
                
                else:
                    setattr(model, var, Var(model.allmeas_times,
                                    model.mixture_components,
                                    bounds=c_bounds,
                                    initialize=1))
        
                    for time, comp in getattr(model, var):
                        if time == model.start_time.value:
                            getattr(model, var)[time, comp].value = v_info[comp].value

        return None
    
    def _add_model_constants(self, model):
        
        if hasattr(self, 'template_constant_data'):
        
            con_info = self.template_constant_data
            
            setattr(model, self.__var.model_constant, Var(con_info.names,
                                                            initialize=con_info.as_dict('value'),
                                                            ))    
        
            for param, obj in getattr(model, self.__var.model_constant).items():
                obj.fix()
          
        return None
    
    
    def _add_model_parameters(self, model):
        """Add the model parameters to the pyomo model"""
        
        if hasattr(self, 'template_parameter_data'):
            
            p_info = self.template_parameter_data
         
            # Initial parameter values
            p_values = p_info.as_dict('value')
            for param, value in p_values.items():
                if value is None:
                    if p_info[param].bounds[0] is not None and p_info[param].bounds[1] is not None:
                        p_values[param] = sum(p_info[param].bounds)/2
                        
            if self._scale_parameters:
                setattr(model, self.__var.model_parameter, Var(model.parameter_names,
                                                               bounds = (0.1, 10),
                                                               initialize=1))
    
            else:
                # print(p_values)
                setattr(model, self.__var.model_parameter,
                        Var(model.parameter_names,
                            initialize=p_values))
    
            # Set the bounds
            p_bounds = p_info.as_dict('bounds')
            for k, v in p_bounds.items():
                factor = 1
                if self._scale_parameters:
                    factor = p_values[k]
                    
                #print(p_info)
    
                if p_info[k].lb is not None:
                    lb = p_info[k].lb/factor
                    getattr(model, self.__var.model_parameter)[k].setlb(lb)
    
                if p_info[k].ub is not None:
                    ub = p_info[k].ub/factor
                    getattr(model, self.__var.model_parameter)[k].setub(ub)
                    
            #for optional smoothing parameters (CS):
            if isinstance(self._smoothparameters, dict) and self._smoothparam_data is not None:
                ps_dict = dict()
                for k in self._smoothparam_data.columns:
                    for c in model.allsmooth_times:
                        ps_dict[c, k] = float(self._smoothparam_data[k][c])
    
                ps_dict2 = dict()
                for p in self._smoothparameters.keys():
                    for t in model.alltime:
                        if t in model.allsmooth_times:
                            ps_dict2[t, p] = float(ps_dict[t, p])
                        else:
                            ps_dict2[t, p] = 0.0
    
                model.Ps = Param(model.alltime, model.smoothparameter_names, initialize=ps_dict2, mutable=True, default=20.)#here just set to some value that is noc
    
            # Fix parameters declared as fixed
            p_fixed = p_info.as_dict('fixed')
            for param, v in p_fixed.items():
                getattr(model, self.__var.model_parameter)[param].fixed = v
        
        return None

    def _add_unwanted_contribution_variables(self, model):
        """Add the bounds for the unwanted contributions, if any"""
        
        if self._qr_bounds is not None:
            qr_bounds = self._qr_bounds
        elif self._qr_bounds is None:
            qr_bounds = None
            
        if self._qr_init is not None:
            qr_init = self._qr_init
        elif self._qr_init is None:
            qr_init = 1.0
            
        model.qr = Var(model.alltime, bounds=qr_bounds, initialize=qr_init)
        
        if self._g_bounds is not None:
            g_bounds = self._g_bounds
        elif self._g_bounds is None:
            g_bounds = None
            
        if self._g_init is not None:
            g_init = self._g_init
        elif self._g_init is None:
            g_init = 0.1
            
        setattr(model, self.__var.unwanted_contribution, Var(model.meas_lambdas, 
                                                             bounds=g_bounds, 
                                                             initialize=g_init))
        
        return None

    # @staticmethod
    def change_time(self, expr_orig, c_mod, new_time, current_model):
        """Method to remove the fixed parameters from the ConcreteModel
        TODO: move to a new expression class
        
        At the moment, this only supports one and two dimensional variables
        """
        expr_new_time = expr_orig
        var_dict = c_mod
        
        for model_var, obj_list in var_dict.items():
            
            if not isinstance(obj_list[1].index(), int):
                old_var = obj_list[1]
                new_var = getattr(current_model, obj_list[0])[new_time, model_var]
        
            else:
                old_var = obj_list[1]
                new_var = getattr(current_model, obj_list[0])[model_var]
        
            expr_new_time = self._update_expression(expr_new_time, old_var, new_var)
    
        return expr_new_time

    @staticmethod
    def _update_expression(expr, replacement_param, change_value):
        """Takes the non-estiambale parameter and replaces it with its intitial
        value
        
        Args:
            expr (pyomo constraint expr): the target ode constraint
            
            replacement_param (str): the non-estimable parameter to replace
            
            change_value (float): initial value for the above parameter
            
        Returns:
            new_expr (pyomo constraint expr): updated constraints with the
                desired parameter replaced with a float
        
        """
        visitor = ReplacementVisitor()
        visitor.change_replacement(change_value)
        visitor.change_suspect(id(replacement_param))
        new_expr = visitor.dfs_postorder_stack(expr)       
        return new_expr

    def _add_model_odes(self, model):
        """Adds the ODE system to the model, if any"""
        
        # if hasattr(self, 'reaction_dict'):
        if isinstance(self._odes, dict):
            
            def rule_odes(m, t, k):
                exprs = self._odes                
               
                if t == m.start_time.value:
                    return Constraint.Skip
                else:
                    if k in m.mixture_components:
                        if k in exprs.keys():
                            deriv_var = f'd{self.__var.concentration_model}dt'
                            final_expr = getattr(m, deriv_var)[t, k] == exprs[k].expression
                            final_expr  = self.change_time(final_expr, self.c_mod, t, m)
                            return final_expr
                        else:
                            return Constraint.Skip
                    else:
                        if k in exprs.keys():
                            deriv_var = f'd{self.__var.state_model}dt'
                            final_expr = getattr(m, deriv_var)[t, k] == exprs[k].expression
                            final_expr  = self.change_time(final_expr, self.c_mod, t, m)
                            return final_expr
                        else:
                            return Constraint.Skip

            setattr(model, self.__var.ode_constraints, Constraint(model.alltime,
                                                                  model.states,
                                                                  rule=rule_odes))
        return None
    
    def _add_algebraic_constraints(self, model):
        """Adds the algebraic constraints the model, if any"""
        
        if self._algebraic_constraints:
            if hasattr(self, 'reaction_dict') or hasattr(self, '_use_alg_dict') and self._use_alg_dict:
                n_alg_eqns = list(self._algebraic_constraints.keys())
                def rule_algebraics(m, t, k):
                    alg_const = self._algebraic_constraints[k].expression
                    alg_var = getattr(m, self.__var.algebraic)[t, k]
                    final_expr = alg_var - alg_const == 0.0
                    final_expr  = self.change_time(final_expr, self.c_mod, t, m)
                    return final_expr
    
            model.algebraic_consts = Constraint(model.alltime,
                                                n_alg_eqns,
                                                rule=rule_algebraics)
        return None
    
    def _add_objective_custom(self, model):
        """Adds the custom objectives, if any and uses the model variable
        UD for the data. This is where the custom data is stored.
        
        """
        def custom_obj(m, t, y):
            exprs = getattr(m, self.__var.user_defined)[t, y] - getattr(m, self.__var.algebraic)[t, y]
            return exprs**2
        
        obj = 0
        var_name = []
        for index, values in model.UD.items():
            if index[1] not in self._custom_objective:
                continue
            if index[1] not in var_name:
                var_name.append(index[1])
            obj += custom_obj(model, index[0], index[1])
            
        model.custom_obj = obj
       
        return None

    def _add_spectral_variables(self, model):
        """Add D and C variables for the spectral data"""
        
        if self._spectral_data is not None:
            s_data_dict = dict()
            for t in model.meas_times:
                for l in model.meas_lambdas:
                    if t in model.meas_times:
                        s_data_dict[t, l] = float(self._spectral_data.loc[t, l])
                    else:
                        s_data_dict[t, l] = float('nan')

            setattr(model, self.__var.spectra_data, Param(model.meas_times,
                                                          model.meas_lambdas,
                                                          domain=Reals,
                                                          initialize=s_data_dict))
            
            setattr(model, self.__var.concentration_spectra, Var(model.meas_times,
                                                                 model.mixture_components,
                                                                 bounds=(0, None),
                                                                 initialize=1))
            
        return None

    def _check_absorbing_species(self, model):
        """Set up the appropriate S depending on absorbing species"""
        
        if self._absorption_data is not None:
            s_dict = dict()
            for k in self._absorption_data.columns:
                for l in self._absorption_data.index:
                    s_dict[l, k] = float(self._absorption_data[k][l])
        else:
            s_dict = 1.0
        
        if self._is_D_deriv == True:
            s_bounds = (None, None)
        else:
            s_bounds = (0.0, None)
        
        if self.has_spectral_data():    
        
            if self._is_non_abs_set:
                self.set_non_absorbing_species(model, self._non_absorbing, check=False)
                setattr(model, self.__var.spectra_species, Var(model.meas_lambdas,
                                                               model.abs_components,
                                                               bounds=s_bounds,
                                                               initialize=s_dict))
            else:
                setattr(model, self.__var.spectra_species, Var(model.meas_lambdas,
                                                               model.mixture_components,
                                                               bounds=s_bounds,
                                                               initialize=s_dict))

            if self._absorption_data is not None:
                for l in model.meas_lambdas:
                    for k in model.mixture_components:
                        getattr(model, self.__var.spectra_species)[l, k].fixed = True
                        
        return None
    
    def _apply_bounds_to_variables(self, model):
        """User specified bounds to certain model variables are added here"""
        
        for bound_set in self._prof_bounds:
            var = bound_set[0]
            component_name = bound_set[1]
            if bound_set[2] is not None:
                bound_time_start = bound_set[2][0]
                bound_time_end = bound_set[2][1]
            upper_bound = bound_set[3][1]
            lower_bound = bound_set[3][0]
            
            for time, comp in getattr(model, var):
                if component_name == comp or component_name is None:
                    if bound_set[2] is not None:
                        if time >= bound_time_start and time < bound_time_end:
                            getattr(model, var)[time, comp].setlb(lower_bound)
                            getattr(model, var)[time, comp].setub(upper_bound)
                    else:
                        getattr(model, var)[time, comp].setlb(lower_bound)
                        getattr(model, var)[time, comp].setub(upper_bound)
                        
        return None

    def _add_model_smoothing_parameters(self, model):
        """Adds smoothing parameters to the model, if any.
        These are optional smoothing parameter values (mutable) read from a file
        
        """
        if self._smoothparam_data is not None:
            model.smoothparameter_names = Set(initialize=self._smoothparameters.keys()) #added for mutable parameters
            model.smooth_param_datatimes = Set(initialize=sorted(self._smoothparam_data.index))
            
            help_dict=dict()
            for k in self._smoothparam_data.index:
                for j in self._smoothparam_data.columns:
                    help_dict[k, j]=float(self._smoothparam_data[j][k])
            
            model.smooth_param_data = Param(self._smoothparam_data.index, sorted(self._smoothparam_data.columns), initialize=help_dict)

        return None

    def _set_up_times(self, model, start_time, end_time):
        
        if self._times is not None:
            if start_time is None:
                start_time = self._times[0]
            if end_time is None:
                end_time = self._times[1]
        else:
            if start_time is None and end_time is None:
                try:    
                    start_time = 0
                    end_time = self.time_span_max
                except:
                    raise ValueError('A model requires a start and end time or a dataset')
        
        list_times = self._meas_times
        m_times = sorted(list_times)
        list_feedtimes = self._feed_times  # For inclusion of discrete feeds CS
        feed_times = sorted(list_feedtimes)  # For inclusion of discrete feeds CS
        m_lambdas = list()
        m_alltimes = m_times
        conc_times = list()

        if self._spectral_data is not None and self._huplc_data is None:
            list_times = list_times.union(set(self._spectral_data.index))
            list_lambdas = list(self._spectral_data.columns)
            m_times = sorted(list_times)
            m_lambdas = sorted(list_lambdas)
            m_alltimes=m_times

        if self._absorption_data is not None:
            if not self._meas_times:
                raise RuntimeError('Need to add measurement times')
            list_times = list(self._meas_times)
            list_lambdas = list(self._absorption_data.index)
            m_times = sorted(list_times)
            m_lambdas = sorted(list_lambdas)

        if self._concentration_data is not None:
            list_times = list_times.union(set(self._concentration_data.index))
            list_concs = list(self._concentration_data.columns)
            m_times = sorted(list_times)
            conc_times = sorted(list_times)
            m_alltimes = sorted(list_times)#has to be changed for including huplc data with conc data!
            m_concs = sorted(list_concs)

        if self._complementary_states_data is not None:
            list_times = list_times.union(set(self._complementary_states_data.index))
            list_comps = list(self._complementary_states_data.columns)
            m_times = sorted(list_times)
            m_alltimes = sorted(list_times)
            m_comps = sorted(list_comps)
            
        if self._custom_data is not None:
            list_times = list_times.union(set(self._custom_data.index))
            list_custom = list(self._custom_data.columns)
            m_times = sorted(list_times)
            m_alltimes = sorted(list_times)
            m_custom = sorted(list_custom)

        #For inclusion of h/uplc data:
        if self._huplc_data is not None and self._spectral_data is None: #added for additional H/UPLC data (CS)
            list_huplctimes = self._huplcmeas_times
            list_huplctimes = list_huplctimes.union(set(self._huplc_data.index))
            list_conDhats = list(self._huplc_data.columns)
            m_huplctimes = sorted(list_huplctimes)

        if self._huplc_data is not None and self._spectral_data is not None:
            list_times = list_times.union(set(self._spectral_data.index))
            list_lambdas = list(self._spectral_data.columns)
            m_times = sorted(list_times)
            m_lambdas = sorted(list_lambdas)
            list_huplctimes = self._huplcmeas_times
            list_huplctimes = list_huplctimes.union(set(self._huplc_data.index))
            list_conDhats = list(self._huplc_data.columns)
            m_huplctimes = sorted(list_huplctimes)
            list_alltimes = list_times.union(list_huplctimes)
            m_alltimes = sorted(list_alltimes)

        #added for optional smoothing CS:
        if self._smoothparam_data is not None:
            if self._huplc_data is not None and self._spectral_data is not None:
                list_alltimes = list_times.union(list_huplctimes)
                m_alltimes = sorted(list_alltimes)
                list_alltimessmooth = list_alltimes.union(set(self._smoothparam_data.index))
                m_allsmoothtimes = sorted(list_alltimessmooth)
            else:
                list_timessmooth = list_times.union(set(self._smoothparam_data.index))
                m_alltimes = m_times
                m_allsmoothtimes = sorted(list_timessmooth)

        if self._huplc_data is not None:
            if m_huplctimes:
                self._check_time_inputs(m_huplctimes, start_time, end_time)
                
        if self._smoothparam_data is not None:
            if m_allsmoothtimes:
                self._check_time_inputs(m_allsmoothtimes, start_time, end_time)
            model.allsmooth_times = Set(initialize=m_allsmoothtimes, ordered=True)

        if m_alltimes:
            self._check_time_inputs(m_alltimes, start_time, end_time)

        self._m_lambdas = m_lambdas

        # For inclusion of discrete feeds CS
        if self._feed_times is not None:
            list_feedtimes = list(self._feed_times)
            feed_times = sorted(list_feedtimes)

        if self._huplc_data is not None:
            model.huplcmeas_times = Set(initialize=m_huplctimes, ordered=True)
            model.huplctime = ContinuousSet(initialize=model.huplcmeas_times,
                                                  bounds=(start_time, end_time))
            
        model.allmeas_times = Set(initialize=m_alltimes, ordered=True) #add for new data structure CS
        model.meas_times = Set(initialize=m_times, ordered=True)
        model.feed_times = Set(initialize=feed_times, ordered=True)  # For inclusion of discrete feeds CS
        model.meas_lambdas = Set(initialize=m_lambdas, ordered=True)
        model.alltime = ContinuousSet(initialize=model.allmeas_times,
                                      bounds=(start_time, end_time)) #add for new data structure CS
        
        model.start_time = Param(initialize=start_time, domain=Reals)
        model.end_time = Param(initialize=end_time, domain=Reals)
        
        return None
    
    @staticmethod
    def _check_time_inputs(time_set, start_time, end_time):
        """Checks the first and last time of a measurement to see if it's in
        the model time bounds
        
        """
        # print(time_set)
        # print(start_time, end_time)
        
        if time_set[0] < start_time:
            raise RuntimeError(f'Measurement time {time_set[0]} not within ({start_time}, {end_time})')
        if time_set[-1] > end_time:
            raise RuntimeError(f'Measurement time {time_set[-1]} not within ({start_time}, {end_time})')
    
        return None
    
    def add_step_vars(self, step_data):
        
        if not hasattr(self, '_step_data'):
            self._step_data = None
        self._step_data = step_data
        return None
    
    def _add_time_steps(self, model):
        
        from kipet.common.model_funs import step_fun
    
        # Add dosing var
        setattr(model, self.__var.dosing_variable, Var(model.alltime,
                                                      [self.__var.dosing_component],
                                                      initialize=0))
            
        def build_step_funs(m, t, k, data):
            
            step = 0
            off_modifier = 0
            for i in range(len(data)):
                if i > 0 and data[i]['switch'] == 'off':
                    off_modifier += 1
                step += step_fun(m, t, num=f'{k}_{i}', **data[i])
                
            return step - off_modifier
        
        def rule_step_function(m, t, k):
            step_const = getattr(m, self.__var.step_variable)[t, k] - build_step_funs(m, t, k, self._step_data[k])    # step_fun(m, t, num=k, **self._step_data[k])
            return step_const == 0.0
        
        if hasattr(self, '_step_data') and len(self._step_data) > 0:
            steps = list(self._step_data.keys())
            
            for step in steps:
                step_info = self._step_data[step]
                num_of_steps = len(step_info)
                step_index = list(range(num_of_steps))
                
            tsc_index = []
            tsc_init = {}
        
            for k, v in self._step_data.items():
                for i, step_dict in enumerate(v):
                    tsc_init[f'{k}_{i}'] = step_dict['time']
                    tsc_index.append(f'{k}_{i}')
            
            setattr(model, self.__var.time_step_change, Var(tsc_index,
                                                           # step_index,
                    bounds=(model.start_time, model.end_time), 
                    initialize=tsc_init,
                    ))
                   
            
            setattr(model, self.__var.step_variable, Var(model.alltime,
                                                         steps,
                                         #                step_index,
                                                         initialize=1.0))
                
            #Constraint to keep track of the step values (for diagnostics)
                
            

            constraint_suffix = 'constraint'
            
            setattr(model, self.__var.step_variable + constraint_suffix,
                    Constraint(model.alltime,
                               steps,
                         #      step_index,
                               rule=rule_step_function))   
            
            # def rule_step_alg(m, t, k, k2, factor):
            #       print(t, k, k2, factor)
            #       step_const = 1/factor*getattr(m, self.__var.algebraic)[t, k] - step_fun(m, t, num=k2, **self._step_data[k2])
            #       return step_const == 0.0

            # if hasattr(self, 'template_algebraic_data'):
            #     alg_steps = self.template_algebraic_data.steps
                
            #     alg_values = {alg.name: alg.value for alg in alg_steps.values()}
            #     alg_set = [alg.name for alg in alg_steps.values()]
            #     alg_step = {alg.name: alg.step for alg in alg_steps.values()}
                           
            #     steps = list(set(steps).difference(alg_set))                  
        
            #     setattr(model, self.__var.algebraic + constraint_suffix,
            #             Constraint(model.alltime,
            #                        alg_set,
            #                        alg_step.values(),
            #                        alg_values.values(),
            #                        rule=rule_step_alg,
            #                        ))
                
                     
                   
        return None
            
    def create_pyomo_model(self, start_time=None, end_time=None):
        """Create a pyomo model.

        This method is the core method for further simulation or optimization studies

        Args:
            start_time (float): initial time considered in the model

            end_time (float): final time considered in the model

        Returns:
            Pyomo ConcreteModel

        """
        if self._spectral_data is not None and self._absorption_data is not None:
            raise RuntimeError('Either add absorption data or spectral data but not both')
        
        pyomo_model = ConcreteModel()

        # Declare Sets
        pyomo_model.mixture_components = Set(initialize=self.template_component_data.names)
        
        if not hasattr(self, 'template_parameter_data'):
            parameter_names = []
        else:
            parameter_names = self.template_parameter_data.names
        pyomo_model.parameter_names = Set(initialize=parameter_names)
        
        if not hasattr(self, 'template_state_data'):
            state_names = []
        else:
            state_names = self.template_state_data.names
        pyomo_model.complementary_states = Set(initialize=state_names)
        
        pyomo_model.states = pyomo_model.mixture_components | pyomo_model.complementary_states
        pyomo_model.measured_data = Set(initialize=self._all_state_data)
        
        # Set up the model time sets and parameters
        self._set_up_times(pyomo_model, start_time, end_time)
        
        # Set up the model by calling the following methods:
        self._add_model_variables(pyomo_model)
        self._add_model_parameters(pyomo_model)
            
        if not hasattr(self, 'template_algebraic_data'):
            alg_names = []
        else:
            alg_names = self.template_algebraic_data.names
        pyomo_model.algebraics = Set(initialize=alg_names)
            
        self._add_algebraic_var(pyomo_model)
        self._add_model_constants(pyomo_model)
        self._add_initial_conditions(pyomo_model)
        self._add_spectral_variables(pyomo_model)
        self._add_model_smoothing_parameters(pyomo_model)
        
        # If unwanted contributions are being handled:
        if self._G_contribution is not None:
            self._add_unwanted_contribution_variables(pyomo_model)

         # Add time step variables
        self._add_time_steps(pyomo_model)

        if not hasattr(self, 'c_mod'):
            self.c_mod = Comp(pyomo_model)
            
        if hasattr(self, 'early_return') and self.early_return:
            return pyomo_model

        # Validate the model before writing constraints
        self._validate_data(pyomo_model)

        # Add constraints
        self._add_model_odes(pyomo_model)
        self._add_algebraic_constraints(pyomo_model)
        
        if self._custom_objective:
            self._add_objective_custom(pyomo_model)
        
        # Check the absorbing species sets
        self._check_absorbing_species(pyomo_model)
        
        # Add bounds, is specified
        self._apply_bounds_to_variables(pyomo_model)
       
        # Add given state standard deviations to the pyomo model
        self._state_sigmas = self.template_component_data.var_variances()
        if hasattr(self, 'template_state_data'):
            self._state_sigmas.update(**self.template_state_data.var_variances())
        
        for k, v in self._state_sigmas.items():
            if v is None:
                self._state_sigmas[k] = 1
                print(f'Warning: No variance provided for model component {k}, it is being set to one')
       
        state_sigmas = {k: v for k, v in self._state_sigmas.items() if k in pyomo_model.measured_data}
        pyomo_model.sigma = Param(pyomo_model.measured_data, domain=Reals, initialize=state_sigmas)
       
        # In case of a second call after known_absorbing has been declared
        if self._huplc_data is not None and self._is_huplc_abs_set:
            self.set_huplc_absorbing_species(pyomo_model, self._huplc_absorbing, self._vol, self._solid_spec, check=False)

        if self._is_known_abs_set:  
            self.set_known_absorbing_species(pyomo_model,
                                             self._known_absorbance,
                                             self._known_absorbance_data,
                                             check=False
                                             )
            
        return pyomo_model

    @property
    def num_parameters(self):
        return len(self._parameters)

    @property
    def num_mixture_components(self):
        return len(self._component_names)

    @property
    def num_complementary_states(self):
        return len(self._complementary_states)

    @property
    def num_algebraics(self):
        return len(self._algebraics)

    @property
    def measurement_times(self):
        return self._meas_times

    @property #added with additional data CS
    def huplcmeasurement_times(self):
        return self._huplcmeas_times

    @property
    def feed_times(self):
        return self._feed_times  # added for feeding points CS

    @num_parameters.setter
    def num_parameters(self):
        raise RuntimeError('Not supported')

    @num_mixture_components.setter
    def num_mixture_components(self):
        raise RuntimeError('Not supported')

    @num_complementary_states.setter
    def num_complementary_states(self):
        raise RuntimeError('Not supported')

    @num_algebraics.setter
    def num_algebraics(self):
        raise RuntimeError('Not supported')

    @measurement_times.setter
    def measurement_times(self):
        raise RuntimeError('Not supported')

    @huplcmeasurement_times.setter
    def huplcmeasurement_times(self):
        raise RuntimeError('Not supported')

    @feed_times.setter
    def feed_times(self):
        raise RuntimeError('Not supported')  # added for feeding points CS

    def has_spectral_data(self):
        return self._spectral_data is not None

    def has_adsorption_data(self):
        return self._absorption_data is not None

    def has_concentration_data(self):
        return self._concentration_data is not None


    def optional_warmstart(self, model):
        if hasattr(model, 'dual') and hasattr(model, 'ipopt_zL_out') and hasattr(model, 'ipopt_zU_out') and hasattr(
                model, 'ipopt_zL_in') and hasattr(model, 'ipopt_zU_in'):
            print('warmstart model components already set up before.')
        else:
            model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
            model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
            model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
            model.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
            model.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)

    # def set_estinit_extra_species(self, model, initextra_est_list, check=True):#added for the estimation of initial conditions which have to be complementary state vars CS
    #     # type: (ConcreteModel, list, bool) -> None
    #     """Sets the non absorbing component of the model.

    #     Args:
    #         initextra_est_list: List of to be estimated initial conditions of complementary state variables.
    #         model: The corresponding model.
    #         check: Safeguard against setting this up twice.
    #     """
    #     if hasattr(model, 'estim_init'):
    #         print("To be estimated initial conditions of complementary state variables were already set up before.")
    #         return

    #     if (self._initextra_est_list and check):
    #         raise RuntimeError('To be estimated initial conditions of complementary state variables have been already set up.')

    #     self._initextraest = initextra_est_list
    #     self._estim_init = True
    #     model.estim_init=Param(initialize=self._estim_init)
        
    #     return None

    # def add_init_extra(self, *args, **kwds):#added for the estimation of initial conditions which have to be complementary state vars CS
    #     """Add a kinetic parameter(s) to the model.

    #     Note:
    #         Plan to change this method add parameters as PYOMO variables

    #         This method tries to mimic a template implementation. Depending
    #         on the argument type it will behave differently

    #     Args:
    #         param1 (str): Species name. Creates a variable species

    #         param1 (list): Species names. Creates a list of variable species

    #         param1 (dict): Map species name(s) to value(s). Creates a fixed parameter(s)

    #     Returns:
    #         None

    #     """
    #     bounds = kwds.pop('bounds', None)
    #     init = kwds.pop('init', None)
    #     self._estim_init=True

    #     if len(args) == 1:
    #         name = args[0]
    #         if isinstance(name, six.string_types):
    #             self._initextraparams[name] = None
    #             if bounds is not None:
    #                 self._initextraparams_bounds[name] = bounds
    #             if init is not None:
    #                 self._initextraparams_init[name] = init
    #         elif isinstance(name, list) or isinstance(name, set):
    #             if bounds is not None:
    #                 if len(bounds) != len(name):
    #                     raise RuntimeError('the list of bounds must be equal to the list of species')
    #             for i, n in enumerate(name):
    #                 self._initextraparams[n] = None
    #                 if bounds is not None:
    #                     self._initextraparams_bounds[n] = bounds[i]
    #                 if init is not None:
    #                     self._initextraparams_init[n] = init[i]
    #         elif isinstance(name, dict):
    #             if bounds is not None:
    #                 if len(bounds) != len(name):
    #                     raise RuntimeError('the list of bounds must be equal to the list of species')
    #             for k, v in name.items():
    #                 self._initextraparams[k] = v
    #                 if bounds is not None:
    #                     self._initextraparams_bounds[k] = bounds[k]
    #                     print(bounds[k])
    #                 if init is not None:
    #                     self._initextraparams_init[k] = init[k]
    #         else:
    #             raise RuntimeError('Species data not supported. Try str')
    #     elif len(args) == 2:
    #         first = args[0]
    #         second = args[1]
    #         if isinstance(first, six.string_types):
    #             self._initextraparams[first] = second
    #             if bounds is not None:
    #                 self._initextraparams_bounds[first] = bounds
    #             if init is not None:
    #                 self._initextraparams_init[first] = init
    #         else:
    #             raise RuntimeError('Species argument not supported. Try str,val')
    #     else:
    #         raise RuntimeError('Species argument not supported. Try str,val')

    def set_non_absorbing_species(self, model, non_abs_list, check=True):
        # type: (ConcreteModel, list, bool) -> None
        """Sets the non absorbing component of the model.

        Args:
            non_abs_list: List of non absorbing components.
            model: The corresponding model.
            check: Safeguard against setting this up twice.
        """
        if hasattr(model, 'non_absorbing'):
            print("non-absorbing species were already set up before.")
            return

        if (self._is_non_abs_set and check):
            raise RuntimeError('Non absorbing species have been already set up.')

        self._is_non_abs_set = True
        self._non_absorbing = non_abs_list
        model.add_component('non_absorbing', Set(initialize=self._non_absorbing))

        C = getattr(model, self.__var.concentration_spectra)
        Z = getattr(model, self.__var.concentration_model)

        times = getattr(model, 'meas_times')
        alltimes = getattr(model, 'allmeas_times')
        allcomps = getattr(model, 'mixture_components')

        model.add_component('fixed_C', ConstraintList())
        new_con = getattr(model, 'fixed_C')

        #############################
        # Exclude non absorbing species from S matrix and create subset Cs of C (CS):
        model.add_component('abs_components_names', Set())
        abscompsnames = [name for name in set(sorted(set(allcomps) - set(self._non_absorbing)))]
        model.add_component('abs_components', Set(initialize=abscompsnames))
        abscomps = getattr(model, 'abs_components')

        model.add_component('Cs', Var(times, abscompsnames))
        Cs = getattr(model, 'Cs')
        model.add_component('matchCsC', ConstraintList())
        matchCsC_con = getattr(model, 'matchCsC')

        for time in alltimes:
            for component in allcomps:
                new_con.add(C[time, component] == Z[time, component])
            for componenta in abscomps:
                if time in times:
                    matchCsC_con.add(Cs[time, componenta] == C[time, componenta])
        ##########################

    def set_known_absorbing_species(self, model, known_abs_list, absorbance_data, check=True):
        # type: (ConcreteModel, list, dataframe, bool) -> None
        """Sets the known absorbance profiles for specific components of the model.

        Args:
            knon_abs_list: List of known species absorbance components.
            model: The corresponding model.
            absorbance_data: the dataframe containing the known component spectra
            check: Safeguard against setting this up twice.
        """
        if hasattr(model, 'known_absorbance'):
            print("species with known absorbance were already set up before.")
            return

        if (self._is_non_abs_set and check):
            raise RuntimeError('Species with known absorbance were already set up before.')

        self._is_known_abs_set = True
        self._known_absorbance = known_abs_list
        self._known_absorbance_data = absorbance_data
        model.add_component('known_absorbance', Set(initialize=self._known_absorbance))
        S = getattr(model, self.__var.spectra_species)
        C = getattr(model, self.__var.concentration_spectra)
        Z = getattr(model, self.__var.concentration_model)
        lambdas = getattr(model, 'meas_lambdas')
        model.known_absorbance_data = self._known_absorbance_data
        for component in self._known_absorbance:
            for l in lambdas:
                S[l, component].set_value(self._known_absorbance_data[component][l])
                S[l, component].fix()
      #########################

        #added for additional huplc data CS
    def set_huplc_absorbing_species(self, model, huplc_abs_list, vol=None, solid_spec=None, check=True):
        # type: (ConcreteModel, list, bool) -> None
        """Sets the non absorbing component of the model.

        Args:
            huplc_abs_list: List of huplc absorbing components.
            vol: reactor volume
            solid_spec: solid species observed with huplc data
            model: The corresponding model.
            check: Safeguard against setting this up twice.
        """
        if hasattr(model, 'huplc_absorbing'):
            print("huplc-absorbing species were already set up before.")
            return

        if (self._is_huplc_abs_set and check):
            raise RuntimeError('huplc-absorbing species have been already set up.')


        self._is_huplc_abs_set = True
        self._vol=vol
        model.add_component('vol', Param(initialize=self._vol))


        self._huplc_absorbing = huplc_abs_list
        model.add_component('huplc_absorbing', Set(initialize=self._huplc_absorbing))
        alltime = getattr(model, 'alltime')
        times = getattr(model, 'huplcmeas_times')
        alltimes = getattr(model, 'allmeas_times')
        timesh = getattr(model, 'huplctime')
        alltime=getattr(model, 'alltime')
        #########
        model.add_component('huplcabs_components_names', Set())
        huplcabscompsnames = [name for name in set(sorted(self._huplc_absorbing))]
        model.add_component('huplcabs_components', Set(initialize=huplcabscompsnames))
        huplcabscomps = getattr(model, 'huplcabs_components')

        self._solid_spec=solid_spec
        if solid_spec is not None:
            solid_spec_arg1keys = [key for key in set(sorted(self._solid_spec.keys()))]
            model.add_component('solid_spec_arg1', Set(initialize=solid_spec_arg1keys))
            solid_spec_arg2keys = [key for key in set(sorted(self._solid_spec.values()))]
            model.add_component('solid_spec_arg2', Set(initialize=solid_spec_arg2keys))
            model.add_component('solid_spec', Var(model.solid_spec_arg1, model.solid_spec_arg2, initialize=self._solid_spec))

        C = getattr(model, self.__var.concentration_spectra)
        Z = getattr(model, self.__var.concentration_model)

        Dhat_dict=dict()
        for k in self._huplc_data.columns:
            for c in self._huplc_data.index:
                Dhat_dict[c, k] = float(self._huplc_data[k][c])


        model.add_component('fixed_Dhat', ConstraintList())
        new_con = getattr(model, 'fixed_Dhat')

        model.add_component('Dhat', Var(times, huplcabscompsnames,initialize=Dhat_dict))
        Dhat = getattr(model, 'Dhat')
        model.add_component('Chat', Var(timesh, huplcabscompsnames,initialize=1))

        Chat = getattr(model, 'Chat')

        model.add_component('matchChatZ', ConstraintList())
        matchChatZ_con = getattr(model, 'matchChatZ')



"""
Development Tools for display purposes

For debugging and other purposes
"""

class Print():
    
    """A print class for debugging - it makes it simpler to turn debugging
    statements on and off during development and later bug finding
    """
    
    def __init__(self, verbose=False):
        
        self.verbose=verbose
        
    def __call__(self, string):
        
        if self.verbose:
            print(string)
        
        return None

if __name__ == '__main__':
    
    pass


"""
This module implements the reduced Hessian parameter selection method outlined
in Chen and Biegler (AIChE 2020).

"""
# Standard library imports
import copy
from string import Template

# Third party imports
import numpy as np
import pandas as pd
from pyomo.environ import (
    Objective,
    SolverFactory,
    Suffix,
    Constraint,
    Param,
    Set,
    )

# KIPET library imports
from kipet.common.objectives import (
    conc_objective,
    comp_objective,
    )
from kipet.common.parameter_handling import (
    check_initial_parameter_values,
    set_scaled_parameter_bounds,
    )
from kipet.common.parameter_ranking import (
    parameter_ratios,
    rank_parameters,
    )

from kipet.core_methods.ParameterEstimator import ParameterEstimator
from kipet.core_methods.ResultsObject import ResultsObject
from kipet.post_model_build.scaling import (
    remove_scaling,
    scale_parameters,
    update_expression,
    )
from kipet.common.ReducedHessian import ReducedHessian

__author__ = 'Kevin McBride'  #: April 2020
    
class EstimationPotential():
    """This class is for estimability analysis. The algorithm here is the one
    presented by Chen and Biegler (accepted AIChE 2020) using the reduced 
    hessian to select the estimable parameters. 

    Attributes:
    
        model_builder (pyomo ConcreteModel): The pyomo model 
    
        simulation_data (pandas.DataFrame): Optional simulation data to use for
            warm starting (Needs testing!)
        
        options (dict): Various options for the esimability algorithm:
        
            nfe (int): The number of finite elements to use in the collocation.
            
            ncp (int): The number of collocation points per finite element.
            
            bound_approach (float): The accepted relative difference for
                determining whether or not a bound is considered active (Step 6).
                
            rho (float): Factor used to determine the lower and upper bounds for
                each parameter in fitting (Step 5).
                
            epsilon (float): The minimum value for parameter values
            
            eta (float): Predetermined cut-off value for accepted std/parameter
                ratios.
                
            max_iter_limit (int): Iteration limits for the estimability algorithm.
            
            verbose (bool): Defaults to False, option to display the progress of
                the algorihm.
                
            debug (bool): Defaults to False, option to ask for user input to 
                proceed during the algorithm.
                
            simulate_start (bool): Option to simulate using the model to 
                warm start the optimization
        
    """

    def __init__(self, model, simulation_data=None, options=None,
                 method='k_aug', solver_opts={}, scaled=True,
                 use_bounds=False, use_duals=False, calc_method='fixed'):
        
        # Options handling
        self.options = {} if options is None else options.copy()
        self._options = options.copy()
        
        self.debug = self._options.pop('debug', False)
        self.verbose = self._options.pop('verbose', False)
        self.nfe = self._options.pop('nfe', 50)
        self.ncp = self._options.pop('ncp', 3)
        self.bound_approach = self._options.pop('bound_approach', 1e-2)
        self.rho = self._options.pop('rho', 10)
        self.epsilon = self._options.pop('epsilon', 1e-16)
        self.eta = self._options.pop('eta', 0.1)
        self.max_iter_limit = self._options.pop('max_iter_limit', 20)
        self.simulate_start = self._options.pop('sim_start', False)
        self.method = method
        self.solver_opts = solver_opts
        self.scaled = scaled
        self.use_bounds = use_bounds
        self.use_duals = use_duals
        self.rh_method = calc_method
        
        # Copy the model
        self.model = copy.deepcopy(model)
        self.simulation_data = simulation_data
        
        self.orig_bounds = None
        if not self.scaled and self.use_bounds:
            self.orig_bounds = {k: (v.lb, v.ub) for k, v in self.model.P.items()}
        
        self.debug = False
        
    def __repr__(self):
        
        repr_str = (f'EstimationPotential({self.model}, simulation_data={"Provided" if self.simulation_data is not None else "None"}, options={self.options})')
        
        return repr_str
        
    def estimate(self):
        """This performs the estimability analysis based on the method
        developed in Chen and Biegler 2020 AIChE...
        
        TODO:
            1. Check method in the last step for clarity in the methodology
            (does not affect the implementation!)
            
            2. Look at the optimization code already in KIPET and see if you
            can use it for anything coded here.
            
            3. Make sure nothing is circular!
        
        Args:
            None
            
        Returns:
            None
            
        """
        bound_check = True     

        flag = False
        step = Template('\n' + '*' * 20 + ' Step $number ' + '*' * 20)
        
        self._model_preparation()
        
        # Step 1
        if self.verbose:
            print(step.substitute(number=1))
            print('Initializing N_pre and N_curr\n')
        
        N_pre = len(self.model.P)
        N_curr = len(self.model.P)
        
        # Step 2
        if self.verbose:
            print(step.substitute(number=2))
            print('Initialize Se and Sf\n')
        
        Se = [parameter for parameter in self.model.P.keys()]
        Sf = []
        
        if self.verbose:
            print(f'\nSe: {Se}')
            print(f'Sf: {Sf}\n')
        
        # Step 3 - Calculate the reduced hessian for the initial stage
        if self.verbose:
            print(step.substitute(number=3))
            print('Calculating the Reduced Hessian for the initial parameter set\n')
        
        reduced_hessian = self._calculate_reduced_hessian(Se)
        
        if self.debug:
            #print(reduced_hessian)
            input("Press Enter to continue...")
    
        # Step 4 - Rank the parameters using Gauss-Jordan Elimination
        if self.verbose:
            print(step.substitute(number=4))
            print('Ranking parameters for estimability and moving to Step 5')
        
        Se, Sf = rank_parameters(self.model, reduced_hessian, Se, epsilon=self.epsilon, eta=self.eta)
        
        if len(Se) >= N_curr:
            number_of_parameters_to_move = len(Se) - N_curr + 1
            for i in range(number_of_parameters_to_move):
                Sf.insert(0, Se.pop()) 
            
            N_pre = len(Se)
            N_curr = len(Se)
       
        if self.verbose:
            print(f'\nThe updated parameter sets are:\nSe: {Se}\nSf: {Sf}')
        
        if self.debug:
            input("Press Enter to continue...")

        # Step 5 - Optimize the estimable parameters
        outer_iteration_counter = 0
        params_counter = 0
        saved_parameters_K = {}
        
        while True:
        
            if outer_iteration_counter > self.max_iter_limit:
                print('Maximum iteration limit reached - check the model!')
                break
            
            inner_iteration_counter = 0
            
            while True:
            
                if inner_iteration_counter > self.max_iter_limit:
                    print('Maximum iteration limit reached - check the model!')
                    break
                
                if self.verbose:
                    print(step.substitute(number=5))
                    print('Optimizing the estimable parameters\n')
                    
                for free_param in Se:
                    self.model.P[free_param].unfix()
                    
                for fixed_param in Sf:
                    if self.scaled:
                        self.model.P[fixed_param].fix(1) # changed from (1) to ()
                    else:
                        self.model.P[fixed_param].fix() # changed from (1) to ()
                        
                ipopt = SolverFactory('ipopt')
                ipopt.solve(self.model, tee=False)#self.verbose)
                
                if self.verbose:
                    self.model.P.display()
                    
                # Step 6 - Check for active bounds
                number_of_active_bounds = 0
                
                if self.verbose:
                    print(step.substitute(number=6))
                    print('Checking for active bounds\n')
                else:
                    None
                
                if bound_check:
                    for key, param in self.model.P.items():
                        if (param.value-param.lb)/param.lb <= self.bound_approach or (param.ub - param.value)/param.value <= self.bound_approach:
                            number_of_active_bounds += 1
                            if self.verbose:
                                print('There is at least one active bound - updating parameters and optimizing again\n')
                            break
                
                else:
                    None
                
                if self.scaled and hasattr(self.model, 'K'):
                    for k, v in self.model.K.items():
                        self.model.K[k] = self.model.K[k] * self.model.P[k].value
                        self.model.P[k].set_value(1)
                        
                    # print(self.model.K.display())
                    # print(self.model.P.display())
                        
                else:
                    set_scaled_parameter_bounds(self.model,
                                                parameter_set=Se,
                                                rho=self.rho,
                                                scaled=self.scaled,
                                                original_bounds=self.orig_bounds)
                    
                if hasattr(self.model, 'K'):
                    param_val_save = 'K'
                else:
                    param_val_save = 'P'
                    
                saved_parameters_K[params_counter] = {k: v.value for k, v in getattr(self.model, param_val_save).items()}
                params_counter += 1
                
                if bound_check:
                    if number_of_active_bounds == 0:
                        if self.verbose:
                            print('There are no active bounds, moving to Step 7')
                        break
                        if self.debug:    
                            input("Press Enter to continue...")
                else:
                    None
                        
                inner_iteration_counter += 1
                if self.debug:
                    input("Press Enter to continue...")
                
            if self.debug:
                self.model.P.display()
                self.model.K.display()
                
            reduced_hessian = self._calculate_reduced_hessian(Se)
            
            # Step 7 - Check the ratios of the parameter std to value
            if self.verbose:
                print(step.substitute(number=7))
                print('Checking the ratios of each parameter in Se')
            
            ratios, eigvals = parameter_ratios(self.model, reduced_hessian, Se, epsilon=self.epsilon)
            
            if self.verbose:
                 print('Ratios:')
                 print(ratios)
            
            ratios_satisfied = max(ratios) < self.eta
        
            if ratios_satisfied:
                if self.verbose:
                    print(f'Step 7 passed, all paramater ratios are less than provided tolerance {self.eta}, moving to Step 10')
                    if self.debug:
                        input("Press Enter to continue...")

                    # Step 10 - Check the current number of free parameters
                    print(step.substitute(number=10))
                    print(f'N_curr = {N_curr}, N_pre = {N_pre}, N_param = {len(self.model.P)}')
                
                if (N_curr == (N_pre - 1)) or (N_curr == len(self.model.P)):
                    if self.verbose:
                        print('Step 10 passed, moving to Step 11, the procedure is finished')
                        print(f'Se: {Se}')
                        print(f'Sf: {Sf}')
                    break
                else:
                    if self.verbose:
                        print('Step 10 failed, moving first parameter from Sf to Se and moving to Step 5')
                    Se.append(Sf.pop(0))
                    N_pre = N_curr
                    N_curr = N_curr + 1
                    if self.debug:
                        input("Press Enter to continue...")
                
            else:
                # Step 8 - Compare number of current estimable parameter with previous iteration
                if self.verbose:
                    print('Step 7 failed, moving on to Step 8')
                    print(step.substitute(number=8))
                    print('Comparing the current number of parameters in Se with the previous number')
                    print(f'N_curr = {N_curr}, N_pre = {N_pre}, N_param = {len(self.model.P)}')
                if N_curr == (N_pre + 1):
                    Sf.insert(0, Se.pop())
                    
                    if self.scaled and hasattr(self.model, 'K'):
                        for k, v in self.model.K.items():
                            self.model.K[k] = saved_parameters_K[params_counter-2][k]
                            self.model.P[k].set_value(1)
                    else:
                        for k, v in self.model.P.items():
                            self.model.P[k].set_value(saved_parameters_K[params_counter-1][k])
                        
                    if self.verbose:
                        print('Step 8 passed, moving to Step 11, reloading last Se the procedure is finished')
                        print(f'Se: {Se}')
                        print(f'Sf: {Sf}')
                    break
                else:
                    # Step 9 - Check the inequality condition given by Eq. 27
                    if self.verbose:
                        print('Step 8 failed, moving to Step 9\n')
                        print(step.substitute(number=9))
                        print('Calculating the inequality from Eq. 27 in Chen and Biegler 2020')
                    if sum(1.0/eigvals) < sum((np.array([self.model.P[v].value for v in Se]))**2)*(self.eta**2) and flag == True:
                        Sf.insert(0, Se.pop())
                        if self.verbose:
                            print('Step 9 passed, moving last parameter from Se into Sf and moving to Step 5')
                            print(f'Se: {Se}')
                            print(f'Sf: {Sf}')
                        N_pre = N_curr
                        N_curr = N_curr - 1

                        if self.debug:
                            input("Press Enter to continue...")
                    else:
                        # Step 2a - Reset the parameter vectors (all in Se)
                        if self.verbose:
                            print('Step 9 failed, moving to Step 2\n')
                            print(step.substitute(number=2))
                            print('Reseting the parameter vectors')
                        flag = True
                        Se = [parameter for parameter in self.model.P.keys()]
                        Sf = []
            
                        if self.debug:
                            input("Press Enter to continue...")
                        
                        if self.verbose:
                            # Step 3a - Recalculate the reduced hessian
                            print(step.substitute(number=3))
                            print('Recalculating the reduced hessian')
                            if self.debug:
                                print(f'Input model:\n')
                                self.model.P.display()
                  
                        reduced_hessian = self._calculate_reduced_hessian(Se)
                        
                        if self.debug:
                            input("Press Enter to continue...")
                        if self.verbose:
                            # Step 4 - Rank the updated parameters using Gauss-Jordan elimination
                            print(step.substitute(number=4))
                            print('Ranking the parameters (limited by N_curr)')
                        
                        Se, Sf = rank_parameters(self.model, reduced_hessian, Se, epsilon=self.epsilon, eta=self.eta)
                        
                        if len(Se) >= N_curr:
                            number_of_parameters_to_move = len(Se) - N_curr + 1
                            for i in range(number_of_parameters_to_move):
                                Sf.insert(0, Se.pop())   
                            
                        N_pre = len(Se)
                        N_curr = len(Se)
                           
                        if self.verbose:
                            print(f'The parameter sets are:\nSe: {Se}\nSf: {Sf}\nN_pre: {N_pre}\nN_curr: {N_curr}')
                        if self.debug:
                            input("Press Enter to continue...")

            outer_iteration_counter += 1                
        
        print(step.substitute(number='Finished'))
        #est_params_str = ', '.join(Se)
        
        self.model.K_vals = saved_parameters_K
        #print(saved_parameters_K)
        
        print(f'The estimable parameters are: {", ".join(Se)}')
        print('\nThe final parameter values are:\n')
        if hasattr(self.model, 'K') and self.model.K is not None:
            self.model.K.pprint()
        else:
            self.model.P.pprint()
            
        results = self._get_results(Se)
            
        return results, self.model
    
    def _get_results(self, Se):
        
        scaled_parameter_var = 'K'
        results = ResultsObject()
        results.estimable_parameters = Se
        
        #results.objective = self.objective_value
        #results.parameter_covariance = self.cov_mat

        # if self._spectra_given:
        #     results.load_from_pyomo_model(self.model,
        #                                   to_load=['Z', 'dZdt', 'X', 'dXdt', 'C', 'S', 'Y'])
        #     if hasattr(self, '_abs_components'):
        #         results.load_from_pyomo_model(self.model,
        #                                       to_load=['Cs'])
        #     if hasattr(self, 'huplc_absorbing'):
        #         results.load_from_pyomo_model(self.model,
        #                                       to_load=['Dhat_bar'])
     
        # elif self._concentration_given:
        results.load_from_pyomo_model(self.model,
                                          to_load=['Z', 'dZdt', 'X', 'U', 'dXdt', 'Cm', 'Y'])
        # else:
        #     raise RuntimeError(
        #         'Must either provide concentration data or spectra in order to solve the parameter estimation problem')

        # if self._spectra_given:
        #     self.compute_D_given_SC(results)

        if hasattr(self.model, scaled_parameter_var): 
            results.P = {name: self.model.P[name].value*getattr(self.model, scaled_parameter_var)[name].value for name in self.model.parameter_names}
        else:
            results.P = {name: self.model.P[name].value for name in self.model.parameter_names}

        # if hasattr(self.model, 'Pinit'):
        #     param_valsinit = dict()
        #     for name in self.model.initparameter_names:
        #         param_valsinit[name] = self.model.init_conditions[name].value
        #     results.Pinit = param_valsinit

        # if self.termination_condition!=None and self.termination_condition!=TerminationCondition.optimal:
        #     raise Exception("The current iteration was unsuccessful.")
        # else:
        #     if self._estimability == True:
        #         return self.hessian, results
        #     else:
        #         return results

        return results
    
    def _model_preparation(self):
        """Helper function that should prepare the models when called from the
        main function. Includes the experimental data, sets the objectives,
        simulates to warm start the models if no data is provided, sets up the
        reduced hessian model with "fake data", and discretizes all models

        """
        if not hasattr(self.model, 'objective'):
            self.model.objective = self._rule_objective(self.model)
        
        # The model needs to be discretized
        model_pe = ParameterEstimator(self.model)
        model_pe.apply_discretization('dae.collocation',
                                      ncp=self.ncp,
                                      nfe=self.nfe,
                                      scheme='LAGRANGE-RADAU')
        
        # Here is where the parameters are scaled
        if self.scaled:
            scale_parameters(self.model)
            set_scaled_parameter_bounds(self.model, rho=self.rho)
        
        else:
            check_initial_parameter_values(self.model)
        
        if self.use_duals:
            self.model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
        
        return None
    
    def _rule_objective(self, model):
        """This function defines the objective function for the estimability
        
        This is equation 5 from Chen and Biegler 2020. It has the following
        form:
            
        .. math::
            \min J = \frac{1}{2}(\mathbf{w}_m - \mathbf{w})^T V_{\mathbf{w}}^{-1}(\mathbf{w}_m - \mathbf{w})
            
        Originally KIPET was designed to only consider concentration data in
        the estimability, but this version now includes complementary states
        such as reactor and cooling temperatures. If complementary state data
        is included in the model, it is detected and included in the objective
        function.
        
        Args:
            model (pyomo.core.base.PyomoModel.ConcreteModel): This is the pyomo
            model instance for the estimability problem.
                
        Returns:
            obj (pyomo.environ.Objective): This returns the objective function
            for the estimability optimization.
        
        """
        obj = 0
        obj += 0.5*conc_objective(model) 
        obj += 0.5*comp_objective(model)
    
        return Objective(expr=obj)

    def _calculate_reduced_hessian(self, Se, verbose=False, **kwargs):
        """This function solves an optimization with very restrictive bounds
        on the paramters in order to get the reduced hessian at fixed 
        conditions
        
        Args:
            Se (list): The current list of estimable parameters.
            
            verbose (bool): Defaults to False, option to show the output from
                the solver (solver option 'tee').
            
        Returns:
            reduced_hessian (np.ndarray): The resulting reduced hessian matrix.
            
        """
        rh_model = copy.deepcopy(self.model)
        
        rh = ReducedHessian(rh_model,
                            parameter_set=Se,
                            rho=self.rho,
                            scaled=self.scaled,
                            param_con_method=self.rh_method,
                            kkt_method=self.method,
                            set_param_bounds = True,
                            )
        
        reduced_hessian = rh.calculate_reduced_hessian(optimize=True)
        
        return reduced_hessian
          
def rhps_method(model, options=None, **kwargs):
    """Reduces a single model using the reduced hessian parameter selection
    method. It takes a pyomo ConcreteModel using P as the parameters to be fit
    and K as the scaled parameter values.
    
    Args:
        model (ConcreteModel): The full model to be reduced
        
        simulation_data (ResultsObject): simulation data for initialization
        
        options (dict): defaults to None, for future option implementations
        
    Returns:
        results (ResultsObject): returns the results from the parameter
            selection and optimization
        reduced_model (ConcreteModel): returns the reduced model with full
            parameter set
    
    """
    simulation_data = kwargs.get('simulation_data', None)
    replace = kwargs.get('replace', True)
    no_scaling = kwargs.get('no_scaling', True)
    method = kwargs.get('method', 'k_aug')
    solver_opts = kwargs.get('solver_opts', {})
    scaled = kwargs.get('scaled', True)
    use_bounds = kwargs.get('use_bounds', False)
    use_duals = kwargs.get('use_duals', False)
    calc_method = kwargs.get('calc_method', 'fixed')
  
    options = options if options is not None else dict()
    orig_bounds = {k: v.bounds for k, v in model.P.items()}
    est_param = EstimationPotential(model,
                                    simulation_data=None,
                                    options=options,
                                    method=method,
                                    solver_opts=solver_opts,
                                    scaled=scaled,
                                    use_bounds=use_bounds,
                                    use_duals=use_duals,
                                    calc_method=calc_method)
    results, reduced_model = est_param.estimate()
    
    if replace:
        reduced_model = replace_non_estimable_parameters(reduced_model, 
                                                         results.estimable_parameters)
    # if no_scaling:
    #     remove_scaling(reduced_model, bounds=orig_bounds)
        
    return results, reduced_model
        
def replace_non_estimable_parameters(model, set_of_est_params):
    """Takes a model and a set of estimable parameters and removes the 
    unestimable parameters from the model by fixing them to their current 
    values in the model
    
    Args:
        model (ConcreteModel): The full model to be reduced
        
        set_of_est_params (set): Parameters found to be estimable
        
    Returns:
        model (ConcreteModel): The model with parameters replaced
        
    """
    all_model_params = set([k for k in model.P.keys()])
    params_to_change = all_model_params.difference(set_of_est_params)
    
    for param in params_to_change:   
        if param in model.P.keys():
            if hasattr(model, 'K'):
                change_value = model.K[param].value
            else:
                change_value = model.P[param].value
        
            for k, v in model.odes.items():
                ep_updated_expr = update_expression(v.body, model.P[param], change_value)
                if hasattr(model, 'K'):
                    ep_updated_expr = update_expression(ep_updated_expr, model.K[param], 1)
                model.odes[k] = ep_updated_expr == 0
    
            model.parameter_names.remove(param)
            del model.P[param]
            if hasattr(model, 'K'):
                del model.K[param]

    return model